"""Market data container for option pricing."""

from __future__ import annotations

from dataclasses import dataclass, replace

from laakhay.quantlab.backend import Backend, get_backend


@dataclass(frozen=True)
class MarketData:
    """Immutable market data bundle: spot, rate, vol."""

    spot: float | object = 100.0
    rate: float | object = 0.0
    vol: float | object = 0.2
    backend: Backend | None = None

    def __post_init__(self) -> None:
        """Initialize backend and validate inputs."""
        # Set default backend if None
        if self.backend is None:
            object.__setattr__(self, "backend", get_backend())

        b = self.backend
        # Note: In quantlab, get_backend returns an AbstractBackend/Backend instance

        # Convert inputs
        object.__setattr__(self, "spot", b.convert(self.spot))
        object.__setattr__(self, "rate", b.convert(self.rate))
        object.__setattr__(self, "vol", b.convert(self.vol))

        # Validate
        if b.any(b.less(self.spot, 0)):
            raise ValueError("spot must be non-negative")
        if b.any(b.less(self.vol, 0)):
            raise ValueError("vol must be non-negative")

    @classmethod
    def create(
        cls,
        *,
        spot: float | object = 100.0,
        rate: float | object = 0.0,
        vol: float | object = 0.2,
        backend: Backend | None = None,
    ) -> MarketData:
        """Create market data with sensible defaults."""
        return cls(spot=spot, rate=rate, vol=vol, backend=backend)

    def with_spot(self, spot: float | object) -> MarketData:
        """Return copy with new spot price."""
        return replace(self, spot=spot)

    def with_rate(self, rate: float | object) -> MarketData:
        """Return copy with new interest rate."""
        return replace(self, rate=rate)

    def with_vol(self, vol: float | object) -> MarketData:
        """Return copy with new volatility."""
        return replace(self, vol=vol)

    @property
    def shape(self) -> tuple:
        """Broadcast shape of market data arrays."""
        b = self.backend
        for arr in (self.spot, self.rate, self.vol):
            if b.ndim(arr) > 0:
                return b.shape(arr)
        return ()

    @property
    def is_scalar(self) -> bool:
        """True if all fields are scalars."""
        b = self.backend
        return not any(b.ndim(a) > 0 for a in (self.spot, self.rate, self.vol))

    def __repr__(self) -> str:
        backend_name = getattr(self.backend, "name", str(self.backend))
        return (
            f"MarketData(spot={self.spot}, rate={self.rate}, "
            f"vol={self.vol}, backend={backend_name})"
        )



"""Base option contract and leg definitions."""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum, auto

from laakhay.quantlab.backend import Backend

from ..utils import infer_backend


class Side(Enum):
    """Represents the side of an option contract."""

    CALL = auto()
    PUT = auto()
    COMPOSITE = auto()


class PositionType(Enum):
    """Represents the type of position in an option contract."""

    LONG = +1
    SHORT = -1


class PayoffType(Enum):
    """Represents the type of payoff for an option contract."""

    EUROPEAN = auto()
    DIGITAL = auto()
    ASIAN = auto()
    BARRIER = auto()
    GEOMETRIC_ASIAN = auto()
    COMPOSITE = auto()
    CASH = auto()


class PositionSide(Enum):
    """Position side for option legs."""

    LONG = 1
    SHORT = -1


@dataclass(frozen=True)
class OptionContract(ABC):
    """Pure contract definition without position/quantity."""

    strike: float | None = None
    expiry: float | None = None

    def __repr__(self) -> str:
        return f"{self.payoff_type.name.lower()}–{self.side.name.lower()}(K={self.strike}, T={self.expiry})"

    @property
    @abstractmethod
    def side(self) -> Side:
        """Which side (call/put/composite)."""
        pass

    @property
    @abstractmethod
    def payoff_type(self) -> PayoffType:
        """Type of payoff."""
        pass

    @property
    def is_analytic(self) -> bool:
        """Has known analytic solution."""
        return self.payoff_type in {
            PayoffType.EUROPEAN,
            PayoffType.DIGITAL,
            PayoffType.GEOMETRIC_ASIAN,
        }

    def is_expired(self, current_time: float = 0.0) -> bool:
        """True if option has expired."""
        if self.expiry is None:
            return False
        return current_time >= self.expiry

    @abstractmethod
    def __call__(self, data, backend: Backend = None):
        """Compute option payoff."""
        pass


@dataclass(frozen=True)
class PathIndependentOption(OptionContract):
    """Base class for path-independent options."""

    def __post_init__(self):
        if self.strike is None or self.strike < 0:
            raise ValueError(f"Invalid strike price: {self.strike}")
        if self.expiry is None or self.expiry < 0:
            raise ValueError(f"Invalid expiry: {self.expiry}")

    @infer_backend
    def __call__(self, data, backend: Backend = None):
        if backend.ndim(data) == 1:
            return self.payoff(data, backend=backend)
        else:
            return self.payoff(data[:, -1], backend=backend)

    @abstractmethod
    def intrinsic_value(self, spot: float, backend: Backend = None) -> float:
        """Intrinsic value at a given spot."""
        pass

    @abstractmethod
    def payoff(self, spot_prices, backend: Backend = None):
        """Payoff for terminal spot price(s)."""
        pass

    def is_itm(self, spot: float) -> bool:
        return spot > self.strike if self.side == Side.CALL else spot < self.strike

    def is_atm(self, spot: float, tol: float = 1e-6) -> bool:
        return abs(spot - self.strike) < tol

    def is_otm(self, spot: float) -> bool:
        return not (self.is_itm(spot) or self.is_atm(spot))

    def moneyness(self, spot: float) -> float:
        if spot <= 0:
            raise ValueError("Spot must be positive")
        return spot / self.strike


@dataclass(frozen=True)
class PathDependentOption(OptionContract):
    """Base class for path-dependent options (Asian, Barrier, etc.)."""

    def __post_init__(self):
        """Validate strike and expiry."""
        if self.strike is None or self.strike < 0:
            raise ValueError(f"Invalid strike price: {self.strike}")
        if self.expiry is None or self.expiry < 0:
            raise ValueError(f"Invalid expiry: {self.expiry}")

    @infer_backend
    def __call__(self, price_paths, backend: Backend = None):
        """Compute payoff from full price paths."""
        paths = backend.ensure_ndim(price_paths, 2, "PathDependentOption requires 2D paths")
        return self._path_payoff(paths, backend=backend)


class BaseLeg(ABC):
    """Base class for all leg types."""

    @property
    @abstractmethod
    def qty(self) -> float:
        pass

    @property
    @abstractmethod
    def is_analytic(self) -> bool:
        pass

    @property
    @abstractmethod
    def payoff_type(self) -> PayoffType:
        pass

    @abstractmethod
    def is_expired(self, current_time: float = 0.0) -> bool:
        pass

    @abstractmethod
    def compute_payoff(self, price_data, backend: Backend = None):
        pass


@dataclass(frozen=True)
class OptionLeg(BaseLeg):
    """Wraps an OptionContract with quantity."""

    contract: OptionContract
    qty: float = 1

    def __post_init__(self):
        if not isinstance(self.contract, OptionContract):
            raise TypeError(f"Expected OptionContract, got {type(self.contract)}")
        if self.qty == 0:
            raise ValueError("Quantity must be non-zero")

    @property
    def side(self) -> PositionSide:
        return PositionSide.LONG if self.qty > 0 else PositionSide.SHORT

    @property
    def is_analytic(self) -> bool:
        return self.contract.is_analytic

    @property
    def payoff_type(self) -> PayoffType:
        return self.contract.payoff_type

    def is_expired(self, current_time: float = 0.0) -> bool:
        return self.contract.is_expired(current_time)

    @infer_backend
    def compute_payoff(self, price_data, backend: Backend = None):
        contract_payoff = self.contract(price_data, backend=backend)
        return backend.mul(contract_payoff, self.qty)

    def __call__(self, S) -> object:
        return self.compute_payoff(S)

    def __repr__(self) -> str:
        side_str = "Long" if self.qty > 0 else "Short"
        qty_str = f"{abs(self.qty)}x " if abs(self.qty) != 1 else ""
        return f"{side_str} {qty_str}{self.contract}"


@dataclass(frozen=True)
class CashLeg(BaseLeg):
    """Wraps a cash amount with expiry."""

    amount: float
    expiry: float

    @property
    def qty(self) -> float:
        return self.amount

    @property
    def is_analytic(self) -> bool:
        return True

    @property
    def payoff_type(self) -> PayoffType:
        return PayoffType.COMPOSITE

    def is_expired(self, current_time: float = 0.0) -> bool:
        return current_time >= self.expiry

    @infer_backend
    def compute_payoff(self, price_data, backend: Backend = None):
        if backend.ndim(price_data) == 0:
            return self.amount
        elif backend.ndim(price_data) == 1:
            return backend.full_like(price_data, self.amount)
        else:
            n_paths = backend.shape(price_data)[0]
            return backend.full((n_paths,), self.amount)

    def __repr__(self) -> str:
        side_str = "Long" if self.amount > 0 else "Short"
        return f"{side_str} {abs(self.amount):.2f} Cash@{self.expiry:.2f}"



from __future__ import annotations

import math
import statistics
from collections.abc import Sequence
from dataclasses import dataclass
from decimal import Decimal

Number = int | float | Decimal


def _to_float_series(values: Sequence[Number]) -> list[float]:
    return [float(v) for v in values]


def compute_period_returns(equity_values: Sequence[Number]) -> list[float]:
    """Compute simple period-over-period returns from equity values."""
    values = _to_float_series(equity_values)
    returns: list[float] = []
    for prev, current in zip(values, values[1:], strict=False):
        if prev <= 0:
            continue
        returns.append((current / prev) - 1.0)
    return returns


def compute_drawdown_series(equity_values: Sequence[Number]) -> list[float]:
    """Compute drawdown series (0 to negative values)."""
    values = _to_float_series(equity_values)
    if not values:
        return []

    peak = values[0]
    drawdowns: list[float] = []
    for value in values:
        if value > peak:
            peak = value
        if peak <= 0:
            drawdowns.append(0.0)
            continue
        drawdowns.append((value / peak) - 1.0)
    return drawdowns


def compute_max_drawdown(equity_values: Sequence[Number]) -> float:
    """Compute maximum drawdown as a negative fraction."""
    drawdowns = compute_drawdown_series(equity_values)
    if not drawdowns:
        return 0.0
    return min(drawdowns)


def compute_annualized_return(
    equity_values: Sequence[Number], periods_per_year: int
) -> float | None:
    """Compute annualized return from start/end equity and number of periods."""
    values = _to_float_series(equity_values)
    if len(values) < 2:
        return None
    start = values[0]
    end = values[-1]
    if start <= 0:
        return None
    periods = len(values) - 1
    if periods <= 0 or periods_per_year <= 0:
        return None
    ratio = end / start
    if ratio <= 0:
        return None
    exponent = periods_per_year / periods
    # Use log-space math for numerical stability on high-frequency/short windows.
    power_log = exponent * math.log(ratio)
    # Prevent overflow in exp for extreme annualization factors.
    if power_log >= 709.0:
        return float("inf")
    if power_log <= -745.0:
        return -1.0
    return math.exp(power_log) - 1.0


def compute_annualized_volatility(
    period_returns: Sequence[float], periods_per_year: int
) -> float | None:
    """Compute annualized volatility using sample standard deviation."""
    if len(period_returns) < 2:
        return None
    stdev = statistics.stdev(period_returns)
    return stdev * math.sqrt(periods_per_year)


def compute_sharpe_ratio(
    period_returns: Sequence[float],
    periods_per_year: int,
    risk_free_rate: float = 0.0,
) -> float | None:
    """Compute annualized Sharpe ratio from period returns."""
    if len(period_returns) < 2:
        return None

    rf_per_period = (1.0 + risk_free_rate) ** (1.0 / periods_per_year) - 1.0
    excess = [r - rf_per_period for r in period_returns]
    stdev = statistics.stdev(excess)
    if stdev == 0:
        return None
    mean_excess = statistics.fmean(excess)
    return (mean_excess / stdev) * math.sqrt(periods_per_year)


@dataclass(frozen=True)
class PerformanceMetrics:
    """Backtest performance summary metrics."""

    total_return: float
    annualized_return: float | None
    annualized_volatility: float | None
    sharpe_ratio: float | None
    max_drawdown: float
    periods: int

    def to_dict(self) -> dict[str, float | int | None]:
        return {
            "total_return": self.total_return,
            "annualized_return": self.annualized_return,
            "annualized_volatility": self.annualized_volatility,
            "sharpe_ratio": self.sharpe_ratio,
            "max_drawdown": self.max_drawdown,
            "periods": self.periods,
        }


def compute_performance_metrics(
    equity_values: Sequence[Number],
    *,
    periods_per_year: int = 252,
    risk_free_rate: float = 0.0,
) -> PerformanceMetrics:
    """Compute standard performance metrics from an equity curve."""
    values = _to_float_series(equity_values)
    if not values:
        values = [0.0]

    start = values[0]
    end = values[-1]
    total_return = ((end / start) - 1.0) if start > 0 else 0.0
    period_returns = compute_period_returns(values)

    return PerformanceMetrics(
        total_return=total_return,
        annualized_return=compute_annualized_return(values, periods_per_year),
        annualized_volatility=compute_annualized_volatility(period_returns, periods_per_year),
        sharpe_ratio=compute_sharpe_ratio(
            period_returns, periods_per_year=periods_per_year, risk_free_rate=risk_free_rate
        ),
        max_drawdown=compute_max_drawdown(values),
        periods=max(0, len(values) - 1),
    )



"""Black-Scholes analytical formulas for European options."""

from ....greeks import Greeks
from ....market import MarketData
from ....options import EuropeanCall, EuropeanPut
from ..calculations import _normal_pdf, compute_d1_d2, compute_discount_factor
from ..registry import PricingFormula


class EuropeanGreeks:
    """Analytical Greeks computation for European options."""

    @staticmethod
    def compute_greeks(option, market: MarketData, d1, d2, is_call: bool) -> Greeks:
        backend = market.backend

        if option.expiry <= 0:
            if is_call:
                is_itm = backend.greater(market.spot, option.strike)
                delta = backend.where(is_itm, 1.0, 0.0)
            else:
                is_itm = backend.less(market.spot, option.strike)
                delta = backend.where(is_itm, -1.0, 0.0)
            zero = backend.zeros_like(delta) if hasattr(backend, "zeros_like") else 0.0
            return Greeks(delta=delta, gamma=zero, vega=zero, theta=zero, rho=zero)

        sqrt_T = backend.sqrt(backend.array(option.expiry))
        discount = compute_discount_factor(option.expiry, market)

        phi_d1 = _normal_pdf(d1, backend)
        N_d1 = backend.norm_cdf(d1)
        N_d2 = backend.norm_cdf(d2)

        if is_call:
            delta = N_d1
            rho = backend.divide(
                backend.mul(backend.mul(option.strike, option.expiry), backend.mul(discount, N_d2)),
                100.0,
            )
            theta_term1 = backend.divide(
                backend.mul(backend.mul(market.spot, phi_d1), market.vol), backend.mul(2.0, sqrt_T)
            )
            theta_term2 = backend.mul(
                backend.mul(market.rate, backend.mul(option.strike, discount)), N_d2
            )
            theta = backend.divide(backend.mul(-1, backend.add(theta_term1, theta_term2)), 365.0)
        else:
            N_minus_d1 = backend.norm_cdf(backend.mul(-1, d1))
            delta = backend.mul(-1, N_minus_d1)
            N_minus_d2 = backend.norm_cdf(backend.mul(-1, d2))
            rho = backend.divide(
                backend.mul(
                    -1,
                    backend.mul(
                        backend.mul(option.strike, option.expiry), backend.mul(discount, N_minus_d2)
                    ),
                ),
                100.0,
            )
            theta_term1 = backend.divide(
                backend.mul(backend.mul(market.spot, phi_d1), market.vol), backend.mul(2.0, sqrt_T)
            )
            theta_term2 = backend.mul(
                backend.mul(market.rate, backend.mul(option.strike, discount)), N_minus_d2
            )
            theta = backend.divide(backend.add(backend.mul(-1, theta_term1), theta_term2), 365.0)

        gamma = backend.divide(phi_d1, backend.mul(backend.mul(market.spot, market.vol), sqrt_T))
        vega = backend.divide(backend.mul(backend.mul(market.spot, phi_d1), sqrt_T), 100.0)

        return Greeks(delta=delta, gamma=gamma, vega=vega, theta=theta, rho=rho)


class EuropeanCallFormula(PricingFormula):
    """European call pricing formula."""

    def price(self, option: EuropeanCall, market: MarketData) -> object:
        backend = market.backend
        d1, d2 = compute_d1_d2(option.strike, option.expiry, market)
        discount = compute_discount_factor(option.expiry, market)
        N_d1 = backend.norm_cdf(d1)
        N_d2 = backend.norm_cdf(d2)
        return backend.add(
            backend.mul(market.spot, N_d1), backend.mul(backend.mul(-option.strike, discount), N_d2)
        )

    def supports(self, option) -> bool:
        return isinstance(option, EuropeanCall)

    def price_with_greeks(self, option: EuropeanCall, market: MarketData) -> tuple[object, Greeks]:
        backend = market.backend
        d1, d2 = compute_d1_d2(option.strike, option.expiry, market)
        discount = compute_discount_factor(option.expiry, market)
        N_d1, N_d2 = backend.norm_cdf(d1), backend.norm_cdf(d2)
        price = backend.add(
            backend.mul(market.spot, N_d1), backend.mul(backend.mul(-option.strike, discount), N_d2)
        )
        greeks = EuropeanGreeks.compute_greeks(option, market, d1, d2, is_call=True)
        return price, greeks


class EuropeanPutFormula(PricingFormula):
    """European put pricing formula."""

    def price(self, option: EuropeanPut, market: MarketData) -> object:
        backend = market.backend
        d1, d2 = compute_d1_d2(option.strike, option.expiry, market)
        discount = compute_discount_factor(option.expiry, market)
        N_minus_d1 = backend.norm_cdf(backend.mul(-1, d1))
        N_minus_d2 = backend.norm_cdf(backend.mul(-1, d2))
        return backend.add(
            backend.mul(backend.mul(option.strike, discount), N_minus_d2),
            backend.mul(-market.spot, N_minus_d1),
        )

    def supports(self, option) -> bool:
        return isinstance(option, EuropeanPut)

    def price_with_greeks(self, option: EuropeanPut, market: MarketData) -> tuple[object, Greeks]:
        backend = market.backend
        d1, d2 = compute_d1_d2(option.strike, option.expiry, market)
        discount = compute_discount_factor(option.expiry, market)
        N_minus_d1, N_minus_d2 = (
            backend.norm_cdf(backend.mul(-1, d1)),
            backend.norm_cdf(backend.mul(-1, d2)),
        )
        price = backend.add(
            backend.mul(backend.mul(option.strike, discount), N_minus_d2),
            backend.mul(-market.spot, N_minus_d1),
        )
        greeks = EuropeanGreeks.compute_greeks(option, market, d1, d2, is_call=False)
        return price, greeks



"""Exceptions for laakhay-quantlab."""


class BackendError(Exception):
    """Base exception for backend errors."""

    pass


class MixedBackendError(BackendError):
    """Raised when arrays from different backends are mixed."""

    pass


class BackendNotFoundError(BackendError):
    """Raised when requested backend is not available."""

    pass


class GuidelineError(Exception):
    """Raised when implementation violates CLAUDE.md guidelines."""

    pass


__all__ = [
    "BackendError",
    "MixedBackendError",
    "BackendNotFoundError",
    "GuidelineError",
]



"""Base classes for stochastic simulations."""

from abc import ABC, abstractmethod

from laakhay.quantlab.backend import Backend


class BaseSampler(ABC):
    """Base sampler for distributions with abstract backend support."""

    def __init__(self, mean: float = 0.0, std: float = 1.0):
        if std <= 0:
            raise ValueError(f"Standard deviation must be positive, got {std}")
        self._mean = float(mean)
        self._std = float(std)

    @abstractmethod
    def _raw_sample(self, shape: tuple[int, ...], backend: Backend) -> object:
        """Draw from standardized distribution using backend."""
        pass

    @abstractmethod
    def sample(
        self,
        shape: tuple[int, ...],
        stratify: bool = False,
        standardize: bool = False,
        backend: Backend | None = None,
    ) -> object:
        """Sample from distribution using backend."""
        pass

    @property
    def mean(self) -> float:
        return self._mean

    @property
    def std(self) -> float:
        return self._std


class BaseSimulation(ABC):
    """Base class for stochastic process simulations."""

    def __init__(self, sampler: BaseSampler):
        self.sampler = sampler

    @abstractmethod
    def generate_paths(
        self, n_paths: int, n_steps: int, expiry: float, backend: Backend | None = None, **kwargs
    ) -> object:
        """Generate stochastic process paths using backend."""
        pass



"""Centralized type definitions for laakhay-quantlab."""

from .backend import (
    ArrayFunction,
    BackendName,
    BackendProtocol,
    BinaryOp,
    GradFunction,
    UnaryOp,
)
from .base import (
    Array,
    ArrayLike,
    Axis,
    JaxArray,
    NumpyNDArray,
    Scalar,
    Shape,
    TorchTensor,
)

__all__ = [
    "Shape",
    "Axis",
    "Array",
    "ArrayLike",
    "Scalar",
    "NumpyNDArray",
    "TorchTensor",
    "JaxArray",
    "BackendName",
    "ArrayFunction",
    "GradFunction",
    "UnaryOp",
    "BinaryOp",
    "BackendProtocol",
]



from abc import ABC, abstractmethod
from collections.abc import Iterator
from datetime import datetime

from laakhay.data.models.bar import Bar
from laakhay.ta.core.dataset import Dataset
from laakhay.ta.core.ohlcv import OHLCV


class DataFeed(ABC):
    """Abstract base class for data feeds."""

    @abstractmethod
    def stream(
        self, start_dt: datetime | None = None, end_dt: datetime | None = None
    ) -> Iterator[Bar]:
        """Yield bars one by one."""
        pass

    @abstractmethod
    def get_history(self, symbol: str, lookback: int) -> Dataset:
        """Get historical data for a symbol up to the current point in stream.

        This is crucial for feeding the Strategy Engine without lookahead bias.
        """
        pass


class MemDataFeed(DataFeed):
    """In-memory grid of bars."""

    def __init__(self, bars: list[Bar], symbol: str, timeframe: str = "1h"):
        """Initialize with a list of bars."""
        # Sort bars by timestamp to be safe
        self.bars = sorted(bars, key=lambda b: b.timestamp)
        self.symbol = symbol
        self.timeframe = timeframe
        self._current_idx = -1

    def stream(
        self, start_dt: datetime | None = None, end_dt: datetime | None = None
    ) -> Iterator[Bar]:
        """Stream bars."""
        self._current_idx = -1
        for i, bar in enumerate(self.bars):
            if start_dt and bar.timestamp < start_dt:
                continue
            if end_dt and bar.timestamp > end_dt:
                break

            self._current_idx = i
            yield bar

    def get_history(self, symbol: str, lookback: int) -> Dataset:
        """Get history up to current_idx."""
        if symbol != self.symbol:
            # In a multi-asset feed, we'd handle this differently
            return Dataset()

        if self._current_idx < 0:
            return Dataset()

        # simple slice: max(0, current - lookback + 1) : current + 1
        start = max(0, self._current_idx - lookback + 1)
        end = self._current_idx + 1

        sliced_bars = self.bars[start:end]
        if not sliced_bars:
            return Dataset()

        # Convert to TA OHLCV
        # Direct construction to handle attribute mismatch (Bar.timestamp vs Bar.ts)
        ohlcv = OHLCV(
            timestamps=tuple(b.timestamp for b in sliced_bars),
            opens=tuple(b.open for b in sliced_bars),
            highs=tuple(b.high for b in sliced_bars),
            lows=tuple(b.low for b in sliced_bars),
            closes=tuple(b.close for b in sliced_bars),
            volumes=tuple(b.volume for b in sliced_bars),
            is_closed=tuple(b.is_closed for b in sliced_bars),
            symbol=self.symbol,
            timeframe=self.timeframe,
        )

        # Wrap in Dataset
        # Note: We create a fresh dataset every step.
        # Optimization: In Phase 2, we can use a rolling buffer or pre-calculated columns.
        ds = Dataset()
        ds.add_series(self.symbol, self.timeframe, ohlcv, source="ohlcv")
        return ds



"""Backend operations."""

from __future__ import annotations

from ..types import Array, Axis
from .registry import get_backend, infer_backend_from_arrays


def ensure_same_backend(*arrays: Array) -> tuple[Array, ...]:
    """Ensure all arrays use same backend."""
    backend_name = infer_backend_from_arrays(*arrays)
    if backend_name is None:
        return arrays

    backend = get_backend(backend_name)
    return tuple(backend.from_numpy(backend.to_numpy(a)) for a in arrays)


class Ops:
    """Backend operations."""

    @staticmethod
    def binary_op(op_name: str, a: Array, b: Array) -> Array:
        backend = get_backend(infer_backend_from_arrays(a, b))
        op = getattr(backend, op_name)
        return op(a, b)

    @staticmethod
    def unary_op(op_name: str, a: Array, **kwargs) -> Array:
        backend = get_backend(infer_backend_from_arrays(a))
        op = getattr(backend, op_name)
        return op(a, **kwargs)

    @staticmethod
    def add(a: Array, b: Array) -> Array:
        return Ops.binary_op("add", a, b)

    @staticmethod
    def sub(a: Array, b: Array) -> Array:
        return Ops.binary_op("sub", a, b)

    @staticmethod
    def mul(a: Array, b: Array) -> Array:
        return Ops.binary_op("mul", a, b)

    @staticmethod
    def div(a: Array, b: Array) -> Array:
        return Ops.binary_op("div", a, b)

    @staticmethod
    def matmul(a: Array, b: Array) -> Array:
        return Ops.binary_op("matmul", a, b)

    @staticmethod
    def sum(a: Array, axis: Axis = None, keepdims: bool = False) -> Array:
        return Ops.unary_op("sum", a, axis=axis, keepdims=keepdims)

    @staticmethod
    def mean(a: Array, axis: Axis = None, keepdims: bool = False) -> Array:
        return Ops.unary_op("mean", a, axis=axis, keepdims=keepdims)

    @staticmethod
    def std(a: Array, axis: Axis = None, keepdims: bool = False) -> Array:
        return Ops.unary_op("std", a, axis=axis, keepdims=keepdims)

    @staticmethod
    def exp(a: Array) -> Array:
        return Ops.unary_op("exp", a)

    @staticmethod
    def log(a: Array) -> Array:
        return Ops.unary_op("log", a)

    @staticmethod
    def sqrt(a: Array) -> Array:
        return Ops.unary_op("sqrt", a)

    @staticmethod
    def abs(a: Array) -> Array:
        return Ops.unary_op("abs", a)

    @staticmethod
    def reshape(a: Array, shape: tuple[int, ...]) -> Array:
        return Ops.unary_op("reshape", a, shape=shape)

    @staticmethod
    def transpose(a: Array, axes: list[int] | None = None) -> Array:
        return Ops.unary_op("transpose", a, axes=axes)

    @staticmethod
    def eq(a: Array, b: Array) -> Array:
        return Ops.binary_op("eq", a, b)

    @staticmethod
    def lt(a: Array, b: Array) -> Array:
        return Ops.binary_op("lt", a, b)

    @staticmethod
    def gt(a: Array, b: Array) -> Array:
        return Ops.binary_op("gt", a, b)

    @staticmethod
    def where(cond: Array, x: Array, y: Array) -> Array:
        backend = get_backend(infer_backend_from_arrays(cond, x, y))
        return backend.where(cond, x, y)

    @staticmethod
    def clip(a: Array, min_val: float | None = None, max_val: float | None = None) -> Array:
        return Ops.unary_op("clip", a, min=min_val, max=max_val)

    @staticmethod
    def max(a: Array, axis: Axis = None, keepdims: bool = False) -> Array:
        return Ops.unary_op("max", a, axis=axis, keepdims=keepdims)

    @staticmethod
    def min(a: Array, axis: Axis = None, keepdims: bool = False) -> Array:
        return Ops.unary_op("min", a, axis=axis, keepdims=keepdims)


add = Ops.add
sub = Ops.sub
mul = Ops.mul
div = Ops.div
matmul = Ops.matmul
exp = Ops.exp
log = Ops.log
sqrt = Ops.sqrt
abs = Ops.abs
sum = Ops.sum
mean = Ops.mean
std = Ops.std
max = Ops.max
min = Ops.min
where = Ops.where
clip = Ops.clip
reshape = Ops.reshape
transpose = Ops.transpose
eq = Ops.eq
lt = Ops.lt
gt = Ops.gt



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import UTC, datetime
from decimal import Decimal
from enum import StrEnum


class OrderSide(StrEnum):
    """Side of the order."""

    BUY = "BUY"
    SELL = "SELL"


class OrderType(StrEnum):
    """Type of the order."""

    MARKET = "MARKET"
    LIMIT = "LIMIT"
    STOP = "STOP"


class OrderStatus(StrEnum):
    """Status of the order."""

    PENDING = "PENDING"
    FILLED = "FILLED"
    CANCELLED = "CANCELLED"
    REJECTED = "REJECTED"


@dataclass
class Order:
    """Order to buy or sell an asset."""

    id: str
    symbol: str
    side: OrderSide
    qty: Decimal
    type: OrderType
    price: Decimal | None = None  # Structurally optional, required for LIMIT
    label: str = ""
    parent_order_id: str | None = None
    oco_group_id: str | None = None
    timestamp: datetime = field(default_factory=lambda: datetime.now(UTC))
    status: OrderStatus = OrderStatus.PENDING

    def __post_init__(self) -> None:
        """Validate order parameters."""
        if self.qty <= 0:
            raise ValueError("Order quantity must be positive")
        if self.type in {OrderType.LIMIT, OrderType.STOP} and (
            self.price is None or self.price <= 0
        ):
            raise ValueError(f"{self.type.value} order must have a positive price")



"""European option contracts."""

from dataclasses import dataclass

from laakhay.quantlab.backend import Backend

from ..utils import infer_backend
from .base import PathIndependentOption, PayoffType, Side


@dataclass(frozen=True)
class EuropeanCall(PathIndependentOption):
    """European call option contract."""

    @property
    def payoff_type(self) -> PayoffType:
        return PayoffType.EUROPEAN

    @property
    def side(self) -> Side:
        return Side.CALL

    @infer_backend
    def payoff(self, spot_prices, backend: Backend = None):
        spot_prices = backend.array(spot_prices)
        strike = backend.array(self.strike)
        zero = backend.array(0.0)
        return backend.maximum(backend.sub(spot_prices, strike), zero)

    def intrinsic_value(self, spot: float, backend: Backend = None) -> float:
        return max(spot - self.strike, 0.0)


@dataclass(frozen=True)
class EuropeanPut(PathIndependentOption):
    """European put option contract."""

    @property
    def payoff_type(self) -> PayoffType:
        return PayoffType.EUROPEAN

    @property
    def side(self) -> Side:
        return Side.PUT

    @infer_backend
    def payoff(self, spot_prices, backend: Backend = None):
        spot_prices = backend.array(spot_prices)
        strike = backend.array(self.strike)
        zero = backend.array(0.0)
        return backend.maximum(backend.sub(strike, spot_prices), zero)

    def intrinsic_value(self, spot: float, backend: Backend = None) -> float:
        return max(self.strike - spot, 0.0)



from __future__ import annotations

from decimal import Decimal
from typing import Any

from ..models import Order, OrderSide, OrderType, Trade


class OrderManager:
    """Manages active orders and matches them against market data."""

    def __init__(self):
        self.active_orders: list[Order] = []
        self.orders_history: list[Order] = []
        self.trades: list[Trade] = []

    def create_order(
        self,
        symbol: str,
        side: OrderSide,
        qty: Decimal,
        type: OrderType,
        price: Decimal | None = None,
        label: str = "",
        parent_order_id: str | None = None,
        oco_group_id: str | None = None,
        timestamp: Any = None,
    ) -> Order:
        """Create and register a new order."""
        order_id = (
            f"{label}-{len(self.orders_history) + 1}"
            if label
            else f"O-{len(self.orders_history) + 1}"
        )

        order = Order(
            id=order_id,
            symbol=symbol,
            side=side,
            qty=qty,
            type=type,
            price=price,
            label=label,
            parent_order_id=parent_order_id,
            oco_group_id=oco_group_id,
            timestamp=timestamp,
        )
        self.active_orders.append(order)
        self.orders_history.append(order)
        return order

    def match_orders(self, bar: Any, symbol: str) -> list[tuple[Order, Decimal]]:
        """Check for order fills against current bar.

        Returns:
            List of (Order, FillPrice) tuples.
        """
        triggered: list[tuple[Order, Decimal]] = []

        for order in self.active_orders:
            if order.symbol != symbol:
                continue

            is_filled = False
            fill_price = order.price

            # LIMIT matching:
            # - LIMIT BUY fills when market trades at or below limit.
            # - LIMIT SELL fills when market trades at or above limit.
            if order.type == OrderType.LIMIT:
                if (order.side == OrderSide.BUY and bar.low <= order.price) or (
                    order.side == OrderSide.SELL and bar.high >= order.price
                ):
                    is_filled = True
                    fill_price = order.price

            # STOP matching:
            # - STOP BUY triggers when market trades at or above stop.
            # - STOP SELL triggers when market trades at or below stop.
            elif (
                order.type == OrderType.STOP
                and order.price
                and (
                    (order.side == OrderSide.SELL and bar.low <= order.price)
                    or (order.side == OrderSide.BUY and bar.high >= order.price)
                )
            ):
                is_filled = True
                fill_price = order.price

            if is_filled:
                triggered.append((order, fill_price))

        # If multiple exit orders trigger inside the same bar (e.g., SL and TP),
        # use a conservative single fill to avoid unrealistic double execution.
        filled: list[tuple[Order, Decimal]] = []
        if triggered:
            if len(triggered) == 1:
                filled = triggered
            else:
                sell_triggers = [item for item in triggered if item[0].side == OrderSide.SELL]
                if sell_triggers:
                    # Conservative for long exits: assume the lowest executable exit price.
                    filled = [min(sell_triggers, key=lambda item: item[1])]
                else:
                    # Conservative for short exits: assume the highest executable exit price.
                    filled = [max(triggered, key=lambda item: item[1])]

        # Remove filled from active
        for order, _ in filled:
            if order in self.active_orders:
                self.active_orders.remove(order)

        return filled

    def cancel_symbol_orders(self, symbol: str) -> None:
        """Cancel all active orders for a symbol."""
        self.active_orders = [o for o in self.active_orders if o.symbol != symbol]

    def find_active_order(self, symbol: str, label: str) -> Order | None:
        """Find an active order by symbol and label."""
        for order in self.active_orders:
            if order.symbol == symbol and order.label == label:
                return order
        return None

    def update_order_price(self, order_id: str, new_price: Decimal) -> bool:
        """Update order price in-place for an active order."""
        if new_price <= 0:
            return False
        for order in self.active_orders:
            if order.id == order_id:
                order.price = new_price
                return True
        return False



from .config import (
    BacktestConfig,
    BreakevenControl,
    FrequencyControl,
    PositionMode,
    TradeDirectionMode,
)
from .engine import BacktestEngine
from .feed import DataFeed, MemDataFeed
from .models import (
    Order,
    OrderSide,
    OrderStatus,
    OrderType,
    Position,
    RiskType,
    Signal,
    SLTPType,
    Trade,
)
from .strategy import Strategy

__all__ = [
    "BacktestEngine",
    "DataFeed",
    "MemDataFeed",
    "Strategy",
    "Order",
    "OrderSide",
    "OrderType",
    "OrderStatus",
    "Trade",
    "Position",
    "Signal",
    "RiskType",
    "SLTPType",
    "BacktestConfig",
    "BreakevenControl",
    "FrequencyControl",
    "PositionMode",
    "TradeDirectionMode",
]



from __future__ import annotations

from laakhay.ta.core.dataset import Dataset
from laakhay.ta.expr.dsl.compiler import ExpressionCompiler
from laakhay.ta.expr.dsl.parser import ExpressionParser
from laakhay.ta.expr.planner.planner import plan_expression
from laakhay.ta.expr.planner.types import PlanResult
from laakhay.ta.expr.runtime.evaluator import RuntimeEvaluator

from ..models import OrderSide, OrderType, RiskType, Signal, SLTPType


class Strategy:
    """Base strategy class using TA expressions."""

    def __init__(
        self,
        entry_signal: str | None = None,
        exit_signal: str | None = None,
        name: str = "Strategy",
        side: str = "long",
        # Advanced Parameters
        risk_size: RiskType | None = None,
        default_sl: SLTPType | None = None,
        default_tp: SLTPType | None = None,
    ) -> None:
        self.name = name
        self.side = side.lower()
        self.entry_plan: PlanResult | None = None
        self.exit_plan: PlanResult | None = None
        self.evaluator = RuntimeEvaluator()
        self.parser = ExpressionParser()
        self.compiler = ExpressionCompiler()

        # Advanced Defaults
        self.default_risk = risk_size
        self.default_sl = default_sl
        self.default_tp = default_tp
        self._required_lookback = 2

        if entry_signal:
            entry_node = self.parser.parse_text(entry_signal)
            compiled_entry = self.compiler.compile(entry_node)
            self.entry_plan = plan_expression(compiled_entry._node)
            self._required_lookback = max(
                self._required_lookback, self._plan_required_lookback(self.entry_plan)
            )

        if exit_signal:
            exit_node = self.parser.parse_text(exit_signal)
            compiled_exit = self.compiler.compile(exit_node)
            self.exit_plan = plan_expression(compiled_exit._node)
            self._required_lookback = max(
                self._required_lookback, self._plan_required_lookback(self.exit_plan)
            )

    def _plan_required_lookback(self, plan: PlanResult | None) -> int:
        if plan is None:
            return 1
        if not plan.requirements.data_requirements:
            return 1
        return max(req.min_lookback for req in plan.requirements.data_requirements)

    def required_lookback(self) -> int:
        """Minimum bars required to evaluate this strategy safely."""
        return max(2, self._required_lookback)

    def prepare(self, symbol: str, timeframe: str) -> None:
        """Prepare strategy for execution."""
        # Could pre-compile here if needed
        pass

    def on_bar(self, dataset: Dataset, symbol: str, timeframe: str) -> list[Signal]:
        """Evaluate strategy for the current bar."""
        signals = {"entry": False, "exit": False}

        # Clear previous cache to ensure fresh evaluation
        self.evaluator.clear_cache()

        # Evaluate Entry
        if self.entry_plan:
            try:
                entry_series = self.evaluator.evaluate(
                    self.entry_plan, dataset, symbol=symbol, timeframe=timeframe
                )
                if entry_series and len(entry_series) > 0:
                    signals["entry"] = bool(entry_series.values[-1])
            except ValueError:
                # Expected during warmup
                pass
            except Exception as e:
                print(f"Strategy Entry Error: {e}")
                pass

        # Evaluate Exit
        if self.exit_plan:
            try:
                exit_series = self.evaluator.evaluate(
                    self.exit_plan, dataset, symbol=symbol, timeframe=timeframe
                )
                if exit_series and len(exit_series) > 0:
                    signals["exit"] = bool(exit_series.values[-1])
            except ValueError:
                pass
            except Exception as e:
                print(f"Strategy Exit Error: {e}")
                pass

        # Convert to Signal Objects
        result_signals = []

        is_long = self.side == "long"
        entry_side = OrderSide.BUY if is_long else OrderSide.SELL
        exit_side = OrderSide.SELL if is_long else OrderSide.BUY

        if signals["entry"]:
            result_signals.append(
                Signal(
                    symbol=symbol,
                    side=entry_side,
                    type=OrderType.MARKET,
                    sl=self.default_sl,
                    tp=self.default_tp,
                    size=self.default_risk,
                )
            )

        if signals["exit"]:
            result_signals.append(Signal(symbol=symbol, side=exit_side, type=OrderType.MARKET))

        return result_signals



"""Greeks data structures with type safety and vectorized operations."""

from __future__ import annotations

from dataclasses import dataclass, fields


@dataclass
class Greeks:
    """Container for option Greeks with vectorized operations."""

    delta: float | object = 0.0
    gamma: float | object = 0.0
    vega: float | object = 0.0
    theta: float | object = 0.0
    rho: float | object = 0.0

    @classmethod
    def from_dict(cls, data: dict[str, float | object]) -> Greeks:
        """Create Greeks from dictionary."""
        kwargs = {f.name: data[f.name] for f in fields(cls) if f.name in data}
        return cls(**kwargs)

    def to_dict(self) -> dict[str, float | object]:
        """Convert to dictionary."""
        return {
            "delta": self.delta,
            "gamma": self.gamma,
            "vega": self.vega,
            "theta": self.theta,
            "rho": self.rho,
        }

    # ------------------------------------------------------------------
    # Mapping-like compatibility layer (legacy tests expect dict interface)
    # ------------------------------------------------------------------

    def keys(self):
        """Return iterable of greek names – behaves like dict.keys()."""
        return ("delta", "gamma", "vega", "theta", "rho")

    def __getitem__(self, item: str):
        if item not in self.keys():
            raise KeyError(item)
        return getattr(self, item)

    def __iter__(self):  # Allows dict(greeks) casting
        yield from self.keys()

    def __len__(self):
        return 5

    def items(self):
        """Yield (name, value) pairs like dict.items()."""
        for k in self.keys():
            yield (k, self[k])

    def __add__(self, other: Greeks) -> Greeks:
        """Portfolio aggregation via addition."""
        if not isinstance(other, Greeks):
            raise TypeError("Can only add Greeks to Greeks")

        return Greeks(
            delta=self._add_values(self.delta, other.delta),
            gamma=self._add_values(self.gamma, other.gamma),
            vega=self._add_values(self.vega, other.vega),
            theta=self._add_values(self.theta, other.theta),
            rho=self._add_values(self.rho, other.rho),
        )

    def __mul__(self, scalar: float | int) -> Greeks:
        """Position sizing via scalar multiplication."""
        if not isinstance(scalar, (int, float)):
            raise TypeError("Can only multiply Greeks by scalar")

        return Greeks(
            delta=self._mul_value(self.delta, scalar),
            gamma=self._mul_value(self.gamma, scalar),
            vega=self._mul_value(self.vega, scalar),
            theta=self._mul_value(self.theta, scalar),
            rho=self._mul_value(self.rho, scalar),
        )

    def __rmul__(self, scalar: float | int) -> Greeks:
        return self.__mul__(scalar)

    def __neg__(self) -> Greeks:
        return self * -1

    def __sub__(self, other: Greeks) -> Greeks:
        return self + (-other)

    def is_scalar(self) -> bool:
        """Check if all Greeks are scalar values."""
        values = [self.delta, self.gamma, self.vega, self.theta, self.rho]
        return all(isinstance(val, (int, float)) for val in values)

    def is_array(self) -> bool:
        """Check if Greeks contain array values."""
        return not self.is_scalar()

    def shape(self) -> tuple[int, ...] | None:
        """Get shape of array Greeks."""
        if self.is_scalar():
            return None
        for value in [self.delta, self.gamma, self.vega, self.theta, self.rho]:
            if hasattr(value, "shape"):
                return tuple(value.shape)
        return None

    def delta_equivalent(self, spot: float | object) -> float | object:
        """Calculate delta-equivalent notional exposure."""
        return self.delta * spot

    def gamma_pnl(self, price_move: float | object) -> float | object:
        """Estimate P&L from gamma exposure."""
        return 0.5 * self.gamma * (price_move**2)

    def vega_pnl(self, vol_change: float | object) -> float | object:
        """Estimate P&L from vega exposure (vol_change in percentage points)."""
        return self.vega * (vol_change * 100)

    def theta_decay(self) -> float | object:
        """Expected 1-day theta decay."""
        return self.theta

    def summary(self) -> str:
        """Concise summary string."""
        if self.is_scalar():
            return (
                f"Greeks(δ={self.delta:.4f}, γ={self.gamma:.6f}, "
                f"ν={self.vega:.4f}, θ={self.theta:.4f}, ρ={self.rho:.4f})"
            )
        return f"Greeks(shape={self.shape()}, arrays=[δ,γ,ν,θ,ρ])"

    @staticmethod
    def _add_values(a: float | object, b: float | object) -> float | object:
        """Add two tensors using appropriate backend operations."""
        return a + b

    @staticmethod
    def _mul_value(value: float | object, scalar: float | int) -> float | object:
        return value * scalar

    def __repr__(self) -> str:
        return self.summary()


def combine_greeks(*greeks_list: Greeks) -> Greeks:
    """Combine multiple Greeks (portfolio aggregation)."""
    if not greeks_list:
        return Greeks()

    result = greeks_list[0]
    for greeks in greeks_list[1:]:
        result = result + greeks
    return result


def scale_greeks(greeks: Greeks, factor: float | int) -> Greeks:
    """Scale Greeks by factor (position sizing)."""
    return greeks * factor


def greeks_from_legs(leg_greeks: list[Greeks], quantities: list[float | int]) -> Greeks:
    """Combine leg Greeks with quantities."""
    if len(leg_greeks) != len(quantities):
        raise ValueError("Greeks and quantities length mismatch")

    scaled = [greeks * qty for greeks, qty in zip(leg_greeks, quantities, strict=False)]
    return combine_greeks(*scaled)



"""Modular Black-Scholes pricer with registry pattern."""

from __future__ import annotations

from laakhay.quantlab.backend import get_backend

from ...greeks import Greeks
from ...market import MarketData
from .formulas.asian import GeometricAsianCallFormula, GeometricAsianPutFormula
from .formulas.barrier import (
    DownAndInCallFormula,
    DownAndInPutFormula,
    DownAndOutCallFormula,
    DownAndOutPutFormula,
    UpAndInCallFormula,
    UpAndInPutFormula,
    UpAndOutCallFormula,
    UpAndOutPutFormula,
)
from .formulas.digital import DigitalCallFormula, DigitalPutFormula
from .formulas.european import EuropeanCallFormula, EuropeanPutFormula
from .registry import pricing_registry


class BlackScholesPricer:
    """Modular Black-Scholes pricer."""

    def __init__(self, backend=None, market: MarketData | None = None):
        self.backend = backend or get_backend()
        self.default_market = market or MarketData(backend=self.backend)
        self._setup_registry()

    def _setup_registry(self):
        pricing_registry.clear()

        from ...options import (
            DigitalCall,
            DigitalPut,
            DownAndInCall,
            DownAndInPut,
            DownAndOutCall,
            DownAndOutPut,
            EuropeanCall,
            EuropeanPut,
            GeometricAsianCall,
            GeometricAsianPut,
            UpAndInCall,
            UpAndInPut,
            UpAndOutCall,
            UpAndOutPut,
        )

        # European
        pricing_registry.register(EuropeanCall, EuropeanCallFormula())
        pricing_registry.register(EuropeanPut, EuropeanPutFormula())

        # Digital
        pricing_registry.register(DigitalCall, DigitalCallFormula())
        pricing_registry.register(DigitalPut, DigitalPutFormula())

        # Asian
        pricing_registry.register(GeometricAsianCall, GeometricAsianCallFormula())
        pricing_registry.register(GeometricAsianPut, GeometricAsianPutFormula())

        # Barrier
        pricing_registry.register(UpAndOutCall, UpAndOutCallFormula)
        pricing_registry.register(UpAndOutPut, UpAndOutPutFormula)
        pricing_registry.register(DownAndOutCall, DownAndOutCallFormula)
        pricing_registry.register(DownAndOutPut, DownAndOutPutFormula)
        pricing_registry.register(UpAndInCall, UpAndInCallFormula)
        pricing_registry.register(UpAndInPut, UpAndInPutFormula)
        pricing_registry.register(DownAndInCall, DownAndInCallFormula)
        pricing_registry.register(DownAndInPut, DownAndInPutFormula)

    def price(self, option, market: MarketData | None = None) -> object:
        market = market or self.default_market
        return pricing_registry.price(option, market)

    def price_with_greeks(self, option, market: MarketData | None = None) -> tuple[object, Greeks]:
        market = market or self.default_market
        return pricing_registry.price_with_greeks(option, market)

    def implied_vol(
        self,
        option,
        price: object,
        market: MarketData,
        vol_bounds: tuple = (1e-8, 5.0),
        tolerance: float = 1e-8,
    ) -> object:
        """Calculate implied volatility using bisection method."""

        def _bisection_solve(objective_func, low, high, tol, max_iter=100):
            """Simple bisection root finder."""
            try:
                obj_low = objective_func(low)
                obj_high = objective_func(high)

                # Check if root exists in interval
                if obj_low * obj_high > 0:
                    # Try expanding bounds
                    if obj_low > 0:  # Need lower vol
                        low = 1e-10
                        obj_low = objective_func(low)
                    else:  # Need higher vol
                        high = 10.0
                        obj_high = objective_func(high)

                    if obj_low * obj_high > 0:
                        return float("nan")

                # Bisection iteration
                for _ in range(max_iter):
                    mid = (low + high) / 2
                    obj_mid = objective_func(mid)

                    if abs(obj_mid) < tol:
                        return mid

                    if obj_low * obj_mid < 0:
                        high = mid
                        obj_high = obj_mid
                    else:
                        low = mid
                        obj_low = obj_mid

                    if high - low < tol:
                        return mid

                return (low + high) / 2

            except (ValueError, RuntimeError):
                return float("nan")

        # Handle scalar case
        if isinstance(price, (int, float)):
            if price <= 0:
                return float("nan")

            def objective(vol_guess: float) -> float:
                return float(self.price(option, market.with_vol(vol_guess))) - price

            return _bisection_solve(objective, vol_bounds[0], vol_bounds[1], tolerance)

        # Handle array case
        price = self.backend.convert(price)
        spot = market.spot
        rate = market.rate

        result = self.backend.zeros_like(price)
        result = self.backend.add(result, float("nan"))  # Initialize with NaN

        # Process each element
        price_flat = self.backend.to_numpy(price).flatten()

        # Handle scalar or array spot/rate
        if self.backend.ndim(spot) == 0:
            spot_flat = [float(spot)] * len(price_flat)
        else:
            spot_flat = self.backend.to_numpy(spot).flatten()

        if self.backend.ndim(rate) == 0:
            rate_flat = [float(rate)] * len(price_flat)
        else:
            rate_flat = self.backend.to_numpy(rate).flatten()

        for i, (p, s, r) in enumerate(zip(price_flat, spot_flat, rate_flat, strict=False)):
            if p <= 0:
                continue

            def objective(vol_guess: float, s=s, r=r, p=p) -> float:
                return float(
                    self.price(
                        option,
                        MarketData(spot=s, rate=r, vol=vol_guess, backend=self.backend),
                    )
                ) - float(p)

            iv = _bisection_solve(objective, vol_bounds[0], vol_bounds[1], tolerance)

            # Convert result to numpy for assignment, then back
            result_np = self.backend.to_numpy(result)
            result_np.flat[i] = iv
            result = self.backend.convert(result_np)

        return result

    def supports(self, option) -> bool:
        return pricing_registry.supports(option)



"""Backend interface."""

from __future__ import annotations

from ..types import Array, Axis, Shape
from .device import Device


class Backend:
    """Backend protocol for array operations."""

    name: str
    pi: float
    e: float
    inf: float
    nan: float

    def name(self) -> str: ...
    def array(self, data: any, dtype: any = None, device: Device | None = None) -> Array: ...
    def convert(self, data: any, dtype: any = None, device: Device | None = None) -> Array: ...
    def randn(self, shape: Shape, dtype: any = None, device: Device | None = None) -> Array: ...
    def is_scalar(self, a: Array) -> bool: ...
    def add(self, a: Array, b: Array) -> Array: ...
    def subtract(self, a: Array, b: Array) -> Array: ...
    def multiply(self, a: Array, b: Array) -> Array: ...
    def divide(self, a: Array, b: Array) -> Array: ...
    def power(self, a: Array, b: Array) -> Array: ...
    def zeros(self, shape: Shape, dtype: any = None, device: Device | None = None) -> Array: ...
    def ones(self, shape: Shape, dtype: any = None, device: Device | None = None) -> Array: ...
    def full(
        self,
        shape: Shape,
        value: float,
        dtype: any = None,
        device: Device | None = None,
    ) -> Array: ...
    def arange(self, *args, dtype: any = None, device: Device | None = None) -> Array: ...
    def linspace(
        self,
        start: float,
        stop: float,
        num: int = 50,
        device: Device | None = None,
    ) -> Array: ...
    def eye(self, n: int, m: int | None = None, device: Device | None = None) -> Array: ...

    def random_key(self, seed: int) -> any: ...
    def random_normal(
        self,
        key: any,
        shape: Shape,
        dtype: any = None,
        device: Device | None = None,
    ) -> Array: ...
    def random_uniform(
        self,
        key: any,
        shape: Shape,
        low: float = 0.0,
        high: float = 1.0,
        dtype: any = None,
        device: Device | None = None,
    ) -> Array: ...
    def random_split(self, key: any, num: int = 2) -> list[any]: ...

    def reshape(self, a: Array, shape: Shape) -> Array: ...
    def transpose(self, a: Array, axes: list[int] | None = None) -> Array: ...
    def concat(self, arrays: list[Array], axis: int = 0) -> Array: ...
    def concatenate(self, arrays: list[Array], axis: int = 0) -> Array: ...
    def stack(self, arrays: list[Array], axis: int = 0) -> Array: ...
    def split(self, a: Array, indices: int | list[int], axis: int = 0) -> list[Array]: ...
    def squeeze(self, a: Array, axis: Axis = None) -> Array: ...
    def expand_dims(self, a: Array, axis: int) -> Array: ...
    def cumsum(self, a: Array, axis: int | None = None) -> Array: ...

    def sum(self, a: Array, axis: Axis = None, keepdims: bool = False) -> Array: ...
    def mean(self, a: Array, axis: Axis = None, keepdims: bool = False) -> Array: ...
    def std(self, a: Array, axis: Axis = None, keepdims: bool = False) -> Array: ...
    def var(self, a: Array, axis: Axis = None, keepdims: bool = False) -> Array: ...
    def min(self, a: Array, axis: Axis = None, keepdims: bool = False) -> Array: ...
    def max(self, a: Array, axis: Axis = None, keepdims: bool = False) -> Array: ...
    def argmin(self, a: Array, axis: Axis = None) -> Array: ...
    def argmax(self, a: Array, axis: Axis = None) -> Array: ...
    def prod(self, a: Array, axis: Axis = None, keepdims: bool = False) -> Array: ...

    def sub(self, a: Array, b: Array) -> Array: ...
    def mul(self, a: Array, b: Array) -> Array: ...
    def div(self, a: Array, b: Array) -> Array: ...
    def pow(self, a: Array, b: Array) -> Array: ...
    def matmul(self, a: Array, b: Array) -> Array: ...
    def dot(self, a: Array, b: Array) -> Array: ...

    def abs(self, a: Array) -> Array: ...
    def sign(self, a: Array) -> Array: ...
    def exp(self, a: Array) -> Array: ...
    def log(self, a: Array) -> Array: ...
    def log10(self, a: Array) -> Array: ...
    def sqrt(self, a: Array) -> Array: ...
    def sin(self, a: Array) -> Array: ...
    def cos(self, a: Array) -> Array: ...
    def tan(self, a: Array) -> Array: ...
    def tanh(self, a: Array) -> Array: ...
    def sinh(self, a: Array) -> Array: ...
    def cosh(self, a: Array) -> Array: ...

    def norm_cdf(self, x: Array) -> Array: ...
    def norm_pdf(self, x: Array) -> Array: ...
    def norm_ppf(self, q: Array) -> Array: ...
    def erf(self, x: Array) -> Array: ...
    def erfc(self, x: Array) -> Array: ...
    def gamma(self, x: Array) -> Array: ...
    def lgamma(self, x: Array) -> Array: ...

    def eq(self, a: Array, b: Array) -> Array: ...
    def ne(self, a: Array, b: Array) -> Array: ...
    def lt(self, a: Array, b: Array) -> Array: ...
    def le(self, a: Array, b: Array) -> Array: ...
    def gt(self, a: Array, b: Array) -> Array: ...
    def ge(self, a: Array, b: Array) -> Array: ...
    def isnan(self, a: Array) -> Array: ...
    def isinf(self, a: Array) -> Array: ...
    def isfinite(self, a: Array) -> Array: ...

    def any(self, a: Array, axis: Axis = None, keepdims: bool = False) -> Array: ...
    def all(self, a: Array, axis: Axis = None, keepdims: bool = False) -> Array: ...
    def logical_not(self, a: Array) -> Array: ...
    def logical_and(self, a: Array, b: Array) -> Array: ...
    def logical_or(self, a: Array, b: Array) -> Array: ...

    def where(self, cond: Array, x: Array, y: Array) -> Array: ...
    def clip(self, a: Array, min: float | None = None, max: float | None = None) -> Array: ...
    def maximum(self, a: Array, b: Array) -> Array: ...
    def minimum(self, a: Array, b: Array) -> Array: ...

    def cholesky(self, a: Array) -> Array: ...
    def solve(self, a: Array, b: Array) -> Array: ...
    def inv(self, a: Array) -> Array: ...
    def det(self, a: Array) -> Array: ...
    def eig(self, a: Array) -> tuple[Array, Array]: ...
    def svd(self, a: Array, full_matrices: bool = True) -> tuple[Array, Array, Array]: ...
    def qr(self, a: Array) -> tuple[Array, Array]: ...
    def norm(self, a: Array, ord: any = None, axis: Axis = None) -> Array: ...

    def shape(self, a: Array) -> Shape: ...
    def dtype(self, a: Array) -> any: ...
    def ndim(self, a: Array) -> int: ...
    def size(self, a: Array) -> int: ...
    def is_array(self, obj: any) -> bool: ...

    def to_numpy(self, a: Array) -> Array: ...
    def from_numpy(self, a: Array) -> Array: ...
    def item(self, a: Array) -> float | int | bool: ...
    def cast(self, a: Array, dtype: any) -> Array: ...
    def copy(self, a: Array) -> Array: ...

    def gather(self, a: Array, indices: Array, axis: int = 0) -> Array: ...
    def scatter(self, a: Array, indices: Array, values: Array, axis: int = 0) -> Array: ...
    def take(self, a: Array, indices: Array, axis: int | None = None) -> Array: ...
    def ensure_ndim(self, a: Array, ndim: int, message: str | None = None) -> Array: ...

    def memory_info(self, device: Device | None = None) -> dict[str, int]: ...
    def clear_cache(self, device: Device | None = None) -> None: ...


class AbstractBackend:
    """Base backend with common implementations."""

    name: str = "abstract"
    pi: float = 3.141592653589793
    e: float = 2.718281828459045
    inf: float = float("inf")
    nan: float = float("nan")

    def full(
        self,
        shape: Shape,
        value: float,
        dtype: any = None,
        device: Device | None = None,
    ) -> Array:
        """Array filled with value."""
        ones = self.ones(shape, dtype=dtype, device=device)
        return self.mul(ones, value)

    def eye(self, n: int, m: int | None = None, device: Device | None = None) -> Array:
        """Identity matrix."""
        m = m or n
        eye = self.zeros((n, m), device=device)
        # TODO: Implement proper eye matrix initialization
        # idx = self.arange(min(n, m), device=device)
        return eye

    def zeros_like(self, a: Array) -> Array:
        """Zeros with same shape/dtype."""
        return self.zeros(self.shape(a), dtype=self.dtype(a), device=self.device(a))

    def ones_like(self, a: Array) -> Array:
        """Ones with same shape/dtype."""
        return self.ones(self.shape(a), dtype=self.dtype(a), device=self.device(a))

    def ensure_ndim(self, a: Array, ndim: int, message: str | None = None) -> Array:
        """Ensure array has exactly ndim dimensions."""
        a = self.array(a)
        if self.ndim(a) != ndim:
            if self.ndim(a) < ndim:
                # Add leading dimensions
                for _ in range(ndim - self.ndim(a)):
                    a = self.expand_dims(a, 0)
            else:
                raise ValueError(message or f"Expected {ndim} dimensions, got {self.ndim(a)}")
        return a

    def convert(self, data: any, dtype: any = None, device: Device | None = None) -> Array:
        """Alias for array() for compatibility."""
        return self.array(data, dtype=dtype, device=device)

    def randn(self, shape: Shape, dtype: any = None, device: Device | None = None) -> Array:
        """Draw samples from standard normal distribution."""
        return self.random_normal(None, shape, dtype=dtype, device=device)

    def is_scalar(self, a: Array) -> bool:
        """Check if array is scalar."""
        return self.ndim(a) == 0

    def concatenate(self, arrays: list[Array], axis: int = 0) -> Array:
        """Alias for concat."""
        return self.concat(arrays, axis=axis)

    def cumsum(self, a: Array, axis: int | None = None) -> Array:
        """Compute cumulative sum."""
        # This will be implemented by child backends
        raise NotImplementedError("cumsum not implemented for this backend")

    # Aliases for convenience
    def add(self, a: Array, b: Array) -> Array: ...  # Handled by child or mixin
    def subtract(self, a: Array, b: Array) -> Array:
        return self.sub(a, b)

    def multiply(self, a: Array, b: Array) -> Array:
        return self.mul(a, b)

    def divide(self, a: Array, b: Array) -> Array:
        return self.div(a, b)

    def power(self, a: Array, b: Array) -> Array:
        return self.pow(a, b)

    def equal(self, a: Array, b: Array) -> Array:
        return self.eq(a, b)

    def not_equal(self, a: Array, b: Array) -> Array:
        return self.ne(a, b)

    def less(self, a: Array, b: Array) -> Array:
        return self.lt(a, b)

    def less_equal(self, a: Array, b: Array) -> Array:
        return self.le(a, b)

    def greater(self, a: Array, b: Array) -> Array:
        return self.gt(a, b)

    def greater_equal(self, a: Array, b: Array) -> Array:
        return self.ge(a, b)

    def copy(self, a: Array) -> Array:
        """Copy array."""
        return self.add(a, self.zeros_like(a))



"""Utilities for pricing module."""

from functools import wraps

from laakhay.quantlab.backend import infer_backend_from_arrays


def infer_backend(func):
    """Decorator to infer backend from arguments if not provided."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        backend = kwargs.get("backend")
        if backend is None:
            # Try to infer from args[1:] (skipping self)
            arrays = [arg for arg in args[1:] if arg is not None]
            # Also check kwargs
            for k, v in kwargs.items():
                if k != "backend" and v is not None:
                    arrays.append(v)

            backend_name = infer_backend_from_arrays(*arrays)
            from laakhay.quantlab.backend import get_backend

            backend = get_backend(backend_name)
            kwargs["backend"] = backend

        return func(*args, **kwargs)

    return wrapper



from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
from datetime import datetime


@dataclass(frozen=True)
class RoundTripTrade:
    """Closed round-trip trade."""

    symbol: str
    direction: str
    qty: float
    entry_price: float
    exit_price: float
    pnl: float
    pnl_pct: float
    holding_seconds: float
    holding_bars: int
    entry_time: datetime | None = None
    exit_time: datetime | None = None
    exit_reason: str | None = None
    sl_price: float | None = None
    tp_price: float | None = None


@dataclass(frozen=True)
class TradeMetrics:
    total_round_trips: int
    win_rate: float | None
    profit_factor: float | None
    expectancy: float | None
    avg_holding_seconds: float | None
    avg_holding_bars: float | None
    avg_win: float | None
    avg_loss: float | None

    def to_dict(self) -> dict[str, float | int | None]:
        return {
            "total_round_trips": self.total_round_trips,
            "win_rate": self.win_rate,
            "profit_factor": self.profit_factor,
            "expectancy": self.expectancy,
            "avg_holding_seconds": self.avg_holding_seconds,
            "avg_holding_bars": self.avg_holding_bars,
            "avg_win": self.avg_win,
            "avg_loss": self.avg_loss,
        }


def compute_trade_metrics(round_trips: Sequence[RoundTripTrade]) -> TradeMetrics:
    """Compute trade-level metrics from closed round trips."""
    total = len(round_trips)
    if total == 0:
        return TradeMetrics(
            total_round_trips=0,
            win_rate=None,
            profit_factor=None,
            expectancy=None,
            avg_holding_seconds=None,
            avg_holding_bars=None,
            avg_win=None,
            avg_loss=None,
        )

    pnls = [trade.pnl for trade in round_trips]
    wins = [p for p in pnls if p > 0]
    losses = [p for p in pnls if p < 0]
    gross_profit = sum(wins)
    gross_loss = abs(sum(losses))

    profit_factor = None if gross_loss == 0 else gross_profit / gross_loss

    avg_win = (sum(wins) / len(wins)) if wins else None
    avg_loss = (sum(losses) / len(losses)) if losses else None

    holding_seconds = [trade.holding_seconds for trade in round_trips]
    holding_bars = [trade.holding_bars for trade in round_trips]

    return TradeMetrics(
        total_round_trips=total,
        win_rate=len(wins) / total,
        profit_factor=profit_factor,
        expectancy=sum(pnls) / total,
        avg_holding_seconds=sum(holding_seconds) / total if holding_seconds else None,
        avg_holding_bars=sum(holding_bars) / total if holding_bars else None,
        avg_win=avg_win,
        avg_loss=avg_loss,
    )



from .performance import (
    PerformanceMetrics,
    compute_drawdown_series,
    compute_max_drawdown,
    compute_performance_metrics,
    compute_period_returns,
    compute_sharpe_ratio,
)
from .timeframe import infer_periods_per_year
from .trades import RoundTripTrade, TradeMetrics, compute_trade_metrics

__all__ = [
    "PerformanceMetrics",
    "RoundTripTrade",
    "TradeMetrics",
    "compute_period_returns",
    "compute_drawdown_series",
    "compute_max_drawdown",
    "compute_sharpe_ratio",
    "compute_performance_metrics",
    "compute_trade_metrics",
    "infer_periods_per_year",
]



"""Registry system for Black-Scholes pricing formulas."""

from abc import ABC, abstractmethod

from ...greeks import Greeks
from ...market import MarketData


class PricingFormula(ABC):
    """Abstract base class for pricing formulas."""

    @abstractmethod
    def price(self, option, market: MarketData) -> object:
        """Calculate option price given market data."""
        pass

    @abstractmethod
    def supports(self, option) -> bool:
        """Check if this formula supports the given option type."""
        pass


class PricingFormulaRegistry:
    """Registry for managing pricing formulas."""

    def __init__(self):
        self._formulas: dict[type, PricingFormula] = {}
        self._priority_formulas: list[PricingFormula] = []

    def register(self, option_type: type, formula: PricingFormula) -> None:
        self._formulas[option_type] = formula

    def register_priority(self, formula: PricingFormula) -> None:
        self._priority_formulas.append(formula)

    def get_formula(self, option) -> PricingFormula:
        option_type = type(option)

        for formula in self._priority_formulas:
            if formula.supports(option):
                return formula

        if option_type in self._formulas:
            return self._formulas[option_type]

        for registered_type, formula in self._formulas.items():
            if isinstance(option, registered_type):
                return formula

        raise ValueError(f"No pricing formula registered for option type: {option_type}")

    def price(self, option, market: MarketData) -> object:
        if isinstance(option, (list, tuple, set)):
            backend = market.backend
            total_price = None
            for single_option in option:
                single_price = self.price(single_option, market)
                if total_price is None:
                    total_price = single_price
                else:
                    total_price = backend.add(total_price, single_price)
            return total_price if total_price is not None else 0.0

        formula = self.get_formula(option)
        return formula.price(option, market)

    def price_with_greeks(self, option, market: MarketData) -> tuple[object, Greeks]:
        if isinstance(option, (list, tuple, set)):
            backend = market.backend
            total_price = None
            total_greeks = None

            for single_option in option:
                single_price, single_greeks = self.price_with_greeks(single_option, market)

                if total_price is None:
                    total_price = single_price
                    total_greeks = single_greeks
                else:
                    total_price = backend.add(total_price, single_price)
                    total_greeks = Greeks(
                        delta=backend.add(total_greeks.delta, single_greeks.delta),
                        gamma=backend.add(total_greeks.gamma, single_greeks.gamma),
                        vega=backend.add(total_greeks.vega, single_greeks.vega),
                        theta=backend.add(total_greeks.theta, single_greeks.theta),
                        rho=backend.add(total_greeks.rho, single_greeks.rho),
                    )

            final_price = total_price if total_price is not None else 0.0
            final_greeks = total_greeks if total_greeks is not None else Greeks()
            return final_price, final_greeks

        formula = self.get_formula(option)

        if hasattr(formula, "price_with_greeks"):
            return formula.price_with_greeks(option, market)

        price = formula.price(option, market)
        backend = market.backend
        zero = backend.zeros_like(price) if hasattr(backend, "zeros_like") else 0.0
        fallback_greeks = Greeks(delta=zero, gamma=zero, vega=zero, theta=zero, rho=zero)

        return price, fallback_greeks

    def supports(self, option) -> bool:
        if isinstance(option, (list, tuple, set)):
            return all(self.supports(single_option) for single_option in option)

        try:
            self.get_formula(option)
            return True
        except ValueError:
            return False

    def clear(self) -> None:
        self._formulas.clear()
        self._priority_formulas.clear()


pricing_registry = PricingFormulaRegistry()



"""Pricing engines and models."""

from .black_scholes import BlackScholesPricer
from .greeks import GreeksEngine
from .monte_carlo import MonteCarloPricer
from .pricer import Pricer, PricingMethod

__all__ = ["Pricer", "PricingMethod", "BlackScholesPricer", "MonteCarloPricer", "GreeksEngine"]



"""Unified pricer interface."""

from __future__ import annotations

from enum import Enum

from laakhay.quantlab.backend import get_backend

from ..greeks import Greeks
from ..market import MarketData
from .black_scholes import BlackScholesPricer
from .monte_carlo import MonteCarloPricer


class PricingMethod(Enum):
    """Available pricing methods."""

    BLACK_SCHOLES = "black_scholes"
    MONTE_CARLO = "monte_carlo"
    AUTO = "auto"


class Pricer:
    """Unified interface for option pricing."""

    def __init__(
        self,
        method: PricingMethod = PricingMethod.AUTO,
        backend=None,
        market: MarketData | None = None,
    ):

        self.backend = get_backend(backend)
        self.method = method
        self.market = market
        self._bs_pricer = BlackScholesPricer(backend=self.backend, market=self.market)
        self._mc_pricer = None  # Lazy init

    @property
    def mc_pricer(self):
        if self._mc_pricer is None:
            self._mc_pricer = MonteCarloPricer(backend=self.backend, market=self.market)
        return self._mc_pricer

    def _select_pricer(self, option):
        """Select appropriate pricer based on method and option."""
        if self.method == PricingMethod.BLACK_SCHOLES:
            return self._bs_pricer
        if self.method == PricingMethod.MONTE_CARLO:
            return self.mc_pricer

        # AUTO logic: try BS only if analytic support exists
        if getattr(option, "is_analytic", False) and self._bs_pricer.supports(option):
            return self._bs_pricer

        return self.mc_pricer

    def price(self, option, market: MarketData | None = None) -> object:
        pricer = self._select_pricer(option)
        return pricer.price(option, market)

    def price_with_greeks(self, option, market: MarketData | None = None) -> tuple[object, Greeks]:
        pricer = self._select_pricer(option)
        if hasattr(pricer, "price_with_greeks"):
            return pricer.price_with_greeks(option, market)

        # Fallback for pricers without analytical Greeks (like MC)
        price = pricer.price(option, market)
        backend = market.backend if market else self.backend
        zero = backend.zeros_like(price) if hasattr(backend, "zeros_like") else 0.0
        return price, Greeks(delta=zero, gamma=zero, vega=zero, theta=zero, rho=zero)



"""Black-Scholes formulas for digital options."""

from ....greeks import Greeks
from ....market import MarketData
from ....options.digital import DigitalCall, DigitalPut
from ..calculations import compute_d1_d2, compute_discount_factor
from ..registry import PricingFormula


def _normal_pdf(x, backend):
    """Calculate normal probability density function φ(x)."""
    sqrt_2pi = backend.sqrt(backend.mul(2.0, backend.pi))
    exp_term = backend.exp(backend.mul(-0.5, backend.power(x, 2)))
    return backend.divide(exp_term, sqrt_2pi)


class DigitalGreeks:
    """Modular Greeks computation for Digital options."""

    @staticmethod
    def compute_greeks(option, market: MarketData, d1, d2, is_call: bool) -> Greeks:
        backend = market.backend

        if option.expiry <= 0:
            zero = backend.zeros_like(market.spot) if hasattr(backend, "zeros_like") else 0.0
            return Greeks(delta=zero, gamma=zero, vega=zero, theta=zero, rho=zero)

        sqrt_T = backend.sqrt(backend.array(option.expiry))
        discount = compute_discount_factor(option.expiry, market)

        phi_d2 = _normal_pdf(d2, backend)
        N_d2 = backend.norm_cdf(d2)

        if is_call:
            # Delta = payout * e^(-rT) * φ(d₂) / (S * vol * √T)
            delta = backend.divide(
                backend.mul(backend.mul(option.payout, discount), phi_d2),
                backend.mul(backend.mul(market.spot, market.vol), sqrt_T),
            )

            # Rho = payout * T * e^(-rT) * N(d₂) / 100
            rho = backend.divide(
                backend.mul(backend.mul(backend.mul(option.payout, option.expiry), discount), N_d2),
                100.0,
            )

            # Theta = payout * e^(-rT) * [r * N(d₂) + φ(d₂) * (r - vol²/2) / (vol * √T)] / 365
            theta_term1 = backend.mul(market.rate, N_d2)
            theta_term2 = backend.divide(
                backend.mul(
                    phi_d2,
                    backend.add(market.rate, backend.mul(-0.5, backend.power(market.vol, 2))),
                ),
                backend.mul(market.vol, sqrt_T),
            )
            theta = backend.divide(
                backend.mul(
                    backend.mul(option.payout, discount), backend.add(theta_term1, theta_term2)
                ),
                365.0,
            )

            # Gamma = -payout * e^(-rT) * φ(d₂) * d₁ / (S² * vol² * T)
            gamma = backend.divide(
                backend.mul(
                    -1, backend.mul(backend.mul(backend.mul(option.payout, discount), phi_d2), d1)
                ),
                backend.mul(
                    backend.mul(backend.power(market.spot, 2), backend.power(market.vol, 2)),
                    option.expiry,
                ),
            )

            # Vega = -payout * e^(-rT) * φ(d₂) * d₁ / (100 * vol)
            vega = backend.divide(
                backend.mul(
                    -1, backend.mul(backend.mul(backend.mul(option.payout, discount), phi_d2), d1)
                ),
                backend.mul(100.0, market.vol),
            )
        else:
            N_minus_d2 = backend.norm_cdf(backend.mul(-1, d2))

            # Delta = -payout * e^(-rT) * φ(d₂) / (S * vol * √T)
            delta = backend.divide(
                backend.mul(-1, backend.mul(backend.mul(option.payout, discount), phi_d2)),
                backend.mul(backend.mul(market.spot, market.vol), sqrt_T),
            )

            # Rho = -payout * T * e^(-rT) * N(-d₂) / 100
            rho = backend.divide(
                backend.mul(
                    -1,
                    backend.mul(
                        backend.mul(backend.mul(option.payout, option.expiry), discount), N_minus_d2
                    ),
                ),
                100.0,
            )

            # Theta = -payout * e^(-rT) * [r * N(-d₂) - φ(d₂) * (r - vol²/2) / (vol * √T)] / 365
            theta_term1 = backend.mul(market.rate, N_minus_d2)
            theta_term2 = backend.divide(
                backend.mul(
                    phi_d2,
                    backend.add(market.rate, backend.mul(-0.5, backend.power(market.vol, 2))),
                ),
                backend.mul(market.vol, sqrt_T),
            )
            theta = backend.divide(
                backend.mul(
                    -1,
                    backend.mul(
                        backend.mul(option.payout, discount),
                        backend.add(theta_term1, backend.mul(-1, theta_term2)),
                    ),
                ),
                365.0,
            )

            # Gamma = payout * e^(-rT) * φ(d₂) * d₁ / (S² * vol² * T)
            gamma = backend.divide(
                backend.mul(backend.mul(backend.mul(option.payout, discount), phi_d2), d1),
                backend.mul(
                    backend.mul(backend.power(market.spot, 2), backend.power(market.vol, 2)),
                    option.expiry,
                ),
            )

            # Vega = payout * e^(-rT) * φ(d₂) * d₁ / (100 * vol)
            vega = backend.divide(
                backend.mul(backend.mul(backend.mul(option.payout, discount), phi_d2), d1),
                backend.mul(100.0, market.vol),
            )

        return Greeks(delta=delta, gamma=gamma, vega=vega, theta=theta, rho=rho)


class DigitalCallFormula(PricingFormula):
    """Digital call pricing formula."""

    def price(self, option: DigitalCall, market: MarketData):
        backend = market.backend
        _, d2 = compute_d1_d2(option.strike, option.expiry, market)
        discount = compute_discount_factor(option.expiry, market)
        N_d2 = backend.norm_cdf(d2)
        return backend.mul(backend.mul(option.payout, discount), N_d2)

    def supports(self, option) -> bool:
        return isinstance(option, DigitalCall)

    def greeks(self, option: DigitalCall, market: MarketData, d1, d2) -> Greeks:
        return DigitalGreeks.compute_greeks(option, market, d1, d2, is_call=True)

    def price_with_greeks(self, option: DigitalCall, market: MarketData):
        d1, d2 = compute_d1_d2(option.strike, option.expiry, market)
        p = self.price(option, market)
        g = self.greeks(option, market, d1, d2)
        return p, g


class DigitalPutFormula(PricingFormula):
    """Digital put pricing formula."""

    def price(self, option: DigitalPut, market: MarketData):
        backend = market.backend
        _, d2 = compute_d1_d2(option.strike, option.expiry, market)
        discount = compute_discount_factor(option.expiry, market)
        N_minus_d2 = backend.norm_cdf(backend.mul(-1, d2))
        return backend.mul(backend.mul(option.payout, discount), N_minus_d2)

    def supports(self, option) -> bool:
        return isinstance(option, DigitalPut)

    def greeks(self, option: DigitalPut, market: MarketData, d1, d2) -> Greeks:
        return DigitalGreeks.compute_greeks(option, market, d1, d2, is_call=False)

    def price_with_greeks(self, option: DigitalPut, market: MarketData):
        d1, d2 = compute_d1_d2(option.strike, option.expiry, market)
        p = self.price(option, market)
        g = self.greeks(option, market, d1, d2)
        return p, g



"""Black-Scholes formulas for barrier options (Reiner & Rubinstein 1991)."""

from __future__ import annotations

from ....greeks import Greeks
from ....market import MarketData
from ....options.barrier import (
    DownAndInCall,
    DownAndInPut,
    DownAndOutCall,
    DownAndOutPut,
    UpAndInCall,
    UpAndInPut,
    UpAndOutCall,
    UpAndOutPut,
)
from ..calculations import compute_barrier_terms
from ..registry import PricingFormula


class BarrierFormula(PricingFormula):
    """Unified barrier option pricing using coefficient matrices."""

    # Coefficient matrices: [A, B, C, D] for [K <= H, K > H] cases
    COEFFICIENTS = {
        "UpAndOutCall": [[1, -1, 1, -1], [0, 0, 0, 0]],
        "UpAndOutPut": [[1, 0, -1, 0], [0, 1, 0, -1]],
        "DownAndOutCall": [[0, 1, 0, -1], [1, 0, -1, 0]],
        "DownAndOutPut": [[0, 0, 0, 0], [1, -1, 1, -1]],
        "UpAndInCall": [[0, 1, -1, 1], [1, 0, 0, 0]],
        "UpAndInPut": [[0, 0, 1, 0], [1, -1, 0, 1]],
        "DownAndInCall": [[1, -1, 0, 1], [0, 0, 1, 0]],
        "DownAndInPut": [[1, 0, 0, 0], [0, 1, -1, 1]],
    }

    def __init__(self, option_type: str, contract_class):
        self.option_type = option_type
        self.contract_class = contract_class
        self.coeffs = self.COEFFICIENTS[option_type]

    def price(self, option, market: MarketData):
        """Price barrier option using coefficient matrix."""
        backend = market.backend

        # Determine phi (Call: 1, Put: -1) and eta (Down: 1, Up: -1)
        phi = 1 if "Call" in self.option_type else -1
        eta = 1 if "Down" in self.option_type else -1

        terms = compute_barrier_terms(option, market, phi=phi, eta=eta)

        # Stack terms: [A, B, C, D]
        terms_array = backend.convert([terms["A"], terms["B"], terms["C"], terms["D"]])

        # Select coefficients based on K <= H condition
        condition = backend.array(option.strike <= option.barrier)
        coeffs_low = backend.array(self.coeffs[0])  # K <= H
        coeffs_high = backend.array(self.coeffs[1])  # K > H

        # selected_coeffs = backend.where(condition, coeffs_low, coeffs_high)
        # We need to handle vectorization if condition is a vector
        if backend.is_scalar(condition):
            selected_coeffs = coeffs_low if condition else coeffs_high
            return backend.sum(backend.mul(selected_coeffs, terms_array), axis=0)
        else:
            # Vectorized selection: this is tricky with the coefficients
            # We'll do it term by term
            price = backend.zeros_like(terms["A"])
            for i in range(4):
                c = backend.where(condition, coeffs_low[i], coeffs_high[i])
                price = backend.add(price, backend.mul(c, terms_array[i]))
            return price

    def supports(self, option) -> bool:
        return isinstance(option, self.contract_class)

    def price_with_greeks(self, option, market: MarketData) -> tuple[object, Greeks]:
        """Price barrier option and compute Greeks (fallback to zeros for now)."""
        price = self.price(option, market)
        # TODO: Implement finite difference greeks if needed
        zero = market.backend.zeros_like(price) if hasattr(market.backend, "zeros_like") else 0.0
        greeks = Greeks(delta=zero, gamma=zero, vega=zero, theta=zero, rho=zero)
        return price, greeks


# Create all formula instances
UpAndOutCallFormula = BarrierFormula("UpAndOutCall", UpAndOutCall)
UpAndOutPutFormula = BarrierFormula("UpAndOutPut", UpAndOutPut)
DownAndOutCallFormula = BarrierFormula("DownAndOutCall", DownAndOutCall)
DownAndOutPutFormula = BarrierFormula("DownAndOutPut", DownAndOutPut)
UpAndInCallFormula = BarrierFormula("UpAndInCall", UpAndInCall)
UpAndInPutFormula = BarrierFormula("UpAndInPut", UpAndInPut)
DownAndInCallFormula = BarrierFormula("DownAndInCall", DownAndInCall)
DownAndInPutFormula = BarrierFormula("DownAndInPut", DownAndInPut)



"""
Composite option contracts and strategy builders.
"""

from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass, field

from laakhay.quantlab.backend import Backend

from ..utils import infer_backend
from .base import BaseLeg, OptionContract, OptionLeg, PayoffType, Side
from .contracts import EuropeanCall, EuropeanPut
from .digital import DigitalCall, DigitalPut


@dataclass(frozen=True)
class CompositeOptionContract(OptionContract):
    """
    Multi-leg option strategy combining individual positions.
    """

    legs: Sequence[BaseLeg] = field(default_factory=list)

    def __post_init__(self):
        """Validate that composite has legs and all are proper BaseLeg types."""
        if not self.legs:
            raise ValueError("CompositeOptionContract requires at least one leg")

        for leg in self.legs:
            if not isinstance(leg, BaseLeg):
                raise TypeError(f"Expected BaseLeg, got {type(leg).__name__}")

    # === Factory Methods ===

    @classmethod
    def from_legs(cls, legs: Sequence[BaseLeg]) -> CompositeOptionContract:
        """Create composite from list of legs."""
        return cls(legs=legs)

    @classmethod
    def combine(
        cls, *composites_or_legs: CompositeOptionContract | BaseLeg | Sequence[BaseLeg]
    ) -> CompositeOptionContract:
        """Combine multiple composites and/or legs into single strategy."""
        all_legs = []

        for item in composites_or_legs:
            if isinstance(item, CompositeOptionContract):
                all_legs.extend(item.legs)
            elif isinstance(item, BaseLeg):
                all_legs.append(item)
            else:
                all_legs.extend(item)

        return cls(legs=all_legs)

    @classmethod
    def merge(
        cls, composite1: CompositeOptionContract, composite2: CompositeOptionContract
    ) -> CompositeOptionContract:
        """Merge two composite strategies into one."""
        return cls(legs=list(composite1.legs) + list(composite2.legs))

    # === String Representations ===

    def __repr__(self) -> str:
        legs_str = ",".join(
            f"{leg.qty:+}×{leg.contract.__class__.__name__}"
            if hasattr(leg, "contract")
            else f"{leg.qty:+}×{leg.__class__.__name__}"
            for leg in self.legs
        )
        return f"Composite({legs_str})"

    def __str__(self) -> str:
        legs_str = ", ".join(str(leg) for leg in self.legs)
        return f"Composite({legs_str})"

    # === Contract Properties ===

    @property
    def side(self) -> Side:
        """Always COMPOSITE side."""
        return Side.COMPOSITE

    @property
    def payoff_type(self) -> PayoffType:
        """COMPOSITE if mixed types, otherwise the unique type."""
        types = {leg.payoff_type for leg in self.legs}
        return types.pop() if len(types) == 1 else PayoffType.COMPOSITE

    @property
    def is_analytic(self) -> bool:
        """True only if ALL legs have analytic solutions."""
        return all(leg.is_analytic for leg in self.legs)

    def is_expired(self, current_time: float = 0.0) -> bool:
        """True if ALL legs have expired."""
        return all(leg.is_expired(current_time) for leg in self.legs)

    # === Payoff Computation ===

    @infer_backend
    def __call__(self, price_data, backend: Backend = None):
        """Compute total payoff across all legs."""
        if not self.legs:
            if backend.ndim(price_data) == 0:
                return 0.0
            elif backend.ndim(price_data) == 1:
                return backend.zeros_like(price_data)
            else:
                n_paths = backend.shape(price_data)[0]
                return backend.zeros((n_paths,))

        total = None
        for leg in self.legs:
            leg_payoff = leg.compute_payoff(price_data, backend=backend)

            total = leg_payoff if total is None else backend.add(total, leg_payoff)

        return total

    def payoff(self, spot_prices, backend: Backend = None):
        """Compute payoff for spot prices."""
        return self(spot_prices, backend=backend)


# === Strategy Builders ===


def vertical_spread(
    lower_strike: float,
    upper_strike: float,
    expiry: float,
    is_call: bool = True,
    is_digital: bool = False,
) -> CompositeOptionContract:
    """Create a vertical spread strategy."""
    if lower_strike >= upper_strike:
        raise ValueError("Lower strike must be less than upper strike")

    if is_digital:
        option_class_lower = DigitalCall if is_call else DigitalPut
        option_class_upper = DigitalCall if is_call else DigitalPut
    else:
        option_class_lower = EuropeanCall if is_call else EuropeanPut
        option_class_upper = EuropeanCall if is_call else EuropeanPut

    legs = [
        OptionLeg(option_class_lower(strike=lower_strike, expiry=expiry), qty=1),
        OptionLeg(option_class_upper(strike=upper_strike, expiry=expiry), qty=-1),
    ]
    return CompositeOptionContract.from_legs(legs)


def calendar_spread(
    strike: float,
    near_expiry: float,
    far_expiry: float,
    is_call: bool = True,
) -> CompositeOptionContract:
    """Create a calendar spread strategy."""
    if near_expiry >= far_expiry:
        raise ValueError("Near expiry must be less than far expiry")

    option_class = EuropeanCall if is_call else EuropeanPut
    legs = [
        OptionLeg(option_class(strike=strike, expiry=near_expiry), qty=-1),
        OptionLeg(option_class(strike=strike, expiry=far_expiry), qty=1),
    ]
    return CompositeOptionContract.from_legs(legs)


def butterfly_spread(
    lower_strike: float,
    middle_strike: float,
    upper_strike: float,
    expiry: float,
    is_call: bool = True,
) -> CompositeOptionContract:
    """Create a butterfly spread strategy."""
    if not (lower_strike < middle_strike < upper_strike):
        raise ValueError("Strikes must be in ascending order")

    option_class = EuropeanCall if is_call else EuropeanPut
    legs = [
        OptionLeg(option_class(strike=lower_strike, expiry=expiry), qty=1),
        OptionLeg(option_class(strike=middle_strike, expiry=expiry), qty=-2),
        OptionLeg(option_class(strike=upper_strike, expiry=expiry), qty=1),
    ]
    return CompositeOptionContract.from_legs(legs)


def straddle(
    strike: float,
    expiry: float,
    is_digital: bool = False,
) -> CompositeOptionContract:
    """Create a straddle strategy."""
    if is_digital:
        call_class, put_class = DigitalCall, DigitalPut
    else:
        call_class, put_class = EuropeanCall, EuropeanPut

    legs = [
        OptionLeg(call_class(strike=strike, expiry=expiry), qty=1),
        OptionLeg(put_class(strike=strike, expiry=expiry), qty=1),
    ]
    return CompositeOptionContract.from_legs(legs)


def strangle(
    call_strike: float,
    put_strike: float,
    expiry: float,
    is_digital: bool = False,
) -> CompositeOptionContract:
    """Create a strangle strategy."""
    if is_digital:
        call_class, put_class = DigitalCall, DigitalPut
    else:
        call_class, put_class = EuropeanCall, EuropeanPut

    legs = [
        OptionLeg(call_class(strike=call_strike, expiry=expiry), qty=1),
        OptionLeg(put_class(strike=put_strike, expiry=expiry), qty=1),
    ]
    return CompositeOptionContract.from_legs(legs)


def iron_condor(
    put_strike_long: float,
    put_strike_short: float,
    call_strike_short: float,
    call_strike_long: float,
    expiry: float,
) -> CompositeOptionContract:
    """Create an iron condor strategy."""
    if not (put_strike_long < put_strike_short < call_strike_short < call_strike_long):
        raise ValueError("Strikes must be in ascending order")

    legs = [
        OptionLeg(EuropeanPut(strike=put_strike_long, expiry=expiry), qty=1),
        OptionLeg(EuropeanPut(strike=put_strike_short, expiry=expiry), qty=-1),
        OptionLeg(EuropeanCall(strike=call_strike_short, expiry=expiry), qty=-1),
        OptionLeg(EuropeanCall(strike=call_strike_long, expiry=expiry), qty=1),
    ]
    return CompositeOptionContract.from_legs(legs)



from laakhay.data.models.bar import Bar

from .order import Order, OrderSide, OrderStatus, OrderType
from .position import Position
from .signal import RiskType, Signal, SLTPType
from .trade import Trade

__all__ = [
    "Bar",
    "Order",
    "OrderSide",
    "OrderStatus",
    "OrderType",
    "Position",
    "Trade",
    "Signal",
    "RiskType",
    "SLTPType",
]



"""Geometric Brownian Motion simulation model."""

from laakhay.quantlab.backend import Backend

from ..utils import infer_backend
from .base import BaseSampler, BaseSimulation
from .samplers import NormalSampler


class GeometricBrownianMotionSimulation(BaseSimulation):
    """Geometric Brownian Motion simulation."""

    def __init__(
        self,
        sampler: NormalSampler | None = None,
        antithetic: bool = False,
        moment_match: bool = False,
        stratify: bool = False,
    ):
        super().__init__(sampler or NormalSampler())
        self.antithetic = antithetic
        self.moment_match = moment_match
        self.stratify = stratify

    @infer_backend
    def generate_shocks(
        self,
        n_paths: int,
        n_steps: int,
        sampler: BaseSampler | None = None,
        backend: Backend | None = None,
    ) -> object:
        """Draw standardized shocks Z ∈ ℝⁿᵖᵃᵗʰˢ×ⁿˢᵗᵉᵖˢ."""
        if sampler is None:
            sampler = self.sampler

        Z = sampler.sample(
            shape=(n_paths, n_steps), stratify=self.stratify, standardize=True, backend=backend
        )

        if self.antithetic and isinstance(sampler, NormalSampler):
            half = n_paths // 2
            Zp = Z[:half]
            Zn = backend.mul(-1, Zp)
            if n_paths % 2 == 0:
                Z = backend.concatenate([Zp, Zn], axis=0)
            else:
                extra = sampler.sample(
                    (1, n_steps), stratify=self.stratify, standardize=True, backend=backend
                )
                Z = backend.concatenate([Zp, Zn, extra], axis=0)

        if self.moment_match:
            m = backend.mean(Z)
            s = backend.std(Z)
            Z = backend.divide(backend.add(Z, backend.mul(-1, m)), backend.add(s, 1e-12))

        return Z

    @infer_backend
    def generate_paths(
        self,
        n_paths: int,
        n_steps: int,
        expiry: float,
        backend: Backend | None = None,
        spot: float = 100.0,
        rate: float = 0.0,
        vol: float = 0.2,
        **kwargs,
    ) -> object:
        """Generate complete GBM paths."""
        shocks = self.generate_shocks(n_paths, n_steps, backend=backend)

        # Prepare parameters as arrays
        spot_arr = backend.array(spot)
        rate_arr = backend.array(rate)
        vol_arr = backend.array(vol)

        # Determine if we need vectorization across parameters
        is_vectorized = (
            backend.ndim(spot_arr) > 0 or backend.ndim(rate_arr) > 0 or backend.ndim(vol_arr) > 0
        )

        dt = expiry / n_steps
        sqrt_dt = backend.sqrt(backend.array(dt))

        # dlog(S) = (r - σ²/2)dt + σdW
        # drift_adj = (r - 0.5 * vol^2) * dt
        vol_sq = backend.power(vol_arr, 2)
        drift_adj = backend.mul(backend.subtract(rate_arr, backend.mul(0.5, vol_sq)), dt)

        if is_vectorized:
            # Expand shocks to (n_paths, n_steps, 1) to broadcast with (V,) params
            shocks_expanded = backend.expand_dims(shocks, axis=-1)
            diffusion = backend.mul(backend.mul(vol_arr, sqrt_dt), shocks_expanded)
            log_returns = backend.add(drift_adj, diffusion)  # (V,) + (P, S, V) -> (P, S, V)

            # Cumulative sums along time axis (axis 1)
            log_prices = backend.cumsum(log_returns, axis=1)  # (P, S, V)

            log_S0 = backend.log(spot_arr)  # (V,)
            # Reshape S0 for concatenation at t=0
            # S0 should be (n_paths, 1, V)
            log_S0_expanded = backend.expand_dims(backend.expand_dims(log_S0, axis=0), axis=0)
            log_S0_full = backend.mul(backend.ones((n_paths, 1, 1)), log_S0_expanded)

            log_prices_with_initial = backend.concatenate(
                [log_S0_full, backend.add(log_S0, log_prices)], axis=1
            )
        else:
            # Scalar case - standard 2D (n_paths, n_steps)
            diffusion = backend.mul(backend.mul(vol_arr, sqrt_dt), shocks)
            log_returns = backend.add(drift_adj, diffusion)
            log_prices = backend.cumsum(log_returns, axis=1)

            log_S0 = backend.log(spot_arr)
            log_S0_col = backend.mul(backend.ones((n_paths, 1)), log_S0)
            log_prices_with_initial = backend.concatenate(
                [log_S0_col, backend.add(log_S0, log_prices)], axis=1
            )

        return backend.exp(log_prices_with_initial)



"""Backend wrapper with autodiff."""

from ..types import Array
from .registry import get_backend as _get_backend
from .registry import infer_backend


class ArrayBackend:
    """Backend wrapper with autodiff support."""

    def __init__(self, backend=None):
        if backend is None:
            self._backend = _get_backend()
        elif isinstance(backend, str):
            self._backend = _get_backend(backend)
        else:
            self._backend = backend

    @property
    def name(self) -> str:
        return self._backend.name

    @property
    def supports_autodiff(self) -> bool:
        """Check autodiff support."""
        return self.name in ["jax", "torch"]

    def __getattr__(self, name):
        return getattr(self._backend, name)

    @staticmethod
    def from_array(array: Array) -> "ArrayBackend":
        """Create backend from array."""
        backend_name = infer_backend(array)
        if backend_name:
            return ArrayBackend(backend_name)
        return ArrayBackend()

    def grad(self, func, argnums=0):
        """Compute gradient."""
        if self.name == "jax":
            import jax

            def scalar_func(*args):
                output = func(*args)
                if hasattr(output, "shape") and output.shape:
                    return output.sum()
                return output

            return jax.grad(scalar_func, argnums=argnums)
        elif self.name == "torch":

            def torch_grad(*args):
                import torch

                tensors = []
                for i, arg in enumerate(args):
                    if i == argnums:
                        t = (
                            arg
                            if isinstance(arg, torch.Tensor)
                            else torch.tensor(arg, dtype=torch.float32)
                        )
                        t.requires_grad_(True)
                        tensors.append(t)
                    else:
                        t = arg if isinstance(arg, torch.Tensor) else torch.tensor(arg)
                        tensors.append(t)

                output = func(*tensors)

                if output.numel() > 1:
                    output = output.sum()

                output.backward()
                return tensors[argnums].grad

            return torch_grad
        else:
            raise RuntimeError(f"Backend {self.name} doesn't support autodiff")

    def value_and_grad(self, func, argnums=0):
        """Compute value and gradient."""
        if self.name == "jax":
            import jax

            def wrapped_func(*args):
                output = func(*args)
                scalar = output.sum() if hasattr(output, "shape") and output.shape else output
                return scalar

            scalar_value_and_grad = jax.value_and_grad(wrapped_func, argnums=argnums)

            def value_and_grad_wrapper(*args):
                original_output = func(*args)
                _, grad = scalar_value_and_grad(*args)
                return original_output, grad

            return value_and_grad_wrapper
        elif self.name == "torch":

            def torch_value_and_grad(*args):
                import torch

                tensors = []
                for i, arg in enumerate(args):
                    if i == argnums:
                        t = (
                            arg
                            if isinstance(arg, torch.Tensor)
                            else torch.tensor(arg, dtype=torch.float32)
                        )
                        t.requires_grad_(True)
                        tensors.append(t)
                    else:
                        t = arg if isinstance(arg, torch.Tensor) else torch.tensor(arg)
                        tensors.append(t)

                output = func(*tensors)
                value = output.detach().clone()

                if output.numel() > 1:
                    output = output.sum()

                output.backward()
                return value, tensors[argnums].grad

            return torch_value_and_grad
        else:
            raise RuntimeError(f"Backend {self.name} doesn't support value_and_grad")

    def vmap(self, func):
        """Vectorize function."""
        if self.name == "jax":
            import jax

            return jax.vmap(func)
        else:

            def vmapped(*args):
                results = [func(x, *args[1:]) for x in args[0]]
                return self._backend.stack(results)

            return vmapped

    def jit(self, func):
        """JIT compile function."""
        if self.name == "jax":
            import jax

            return jax.jit(func)
        else:
            return func


def backend(name=None) -> ArrayBackend:
    """Get backend instance."""
    return ArrayBackend(name)


def active_backend() -> ArrayBackend:
    """Get active backend."""
    return ArrayBackend()



from __future__ import annotations

from dataclasses import dataclass
from decimal import Decimal

from .order import OrderSide, OrderType

# Type Definitions
RiskType = float | Decimal | str  # e.g. 0.01, Decimal("100"), "1%"
SLTPType = Decimal | str  # e.g. Decimal("50000"), "2%"


@dataclass
class Signal:
    """Trading signal with advanced parameters."""

    symbol: str
    side: OrderSide
    type: OrderType
    price: Decimal | None = None  # Required for LIMIT/STOP

    # Risk Management (Overrides Strategy defaults)
    sl: SLTPType | None = None
    tp: SLTPType | None = None

    # Position Sizing (Overrides Strategy defaults)
    size: RiskType | None = None



"""Black-Scholes formulas for geometric Asian options."""

from ....greeks import Greeks
from ....market import MarketData
from ....options.asian import GeometricAsianCall, GeometricAsianPut
from ..calculations import compute_asian_adjusted_params, compute_d1_d2, compute_discount_factor
from ..registry import PricingFormula


def _normal_pdf(x, backend):
    """Calculate normal probability density function φ(x)."""
    sqrt_2pi = backend.sqrt(backend.mul(2.0, backend.pi))
    exp_term = backend.exp(backend.mul(-0.5, backend.power(x, 2)))
    return backend.divide(exp_term, sqrt_2pi)


class AsianGreeks:
    """Modular Greeks computation for Geometric Asian options."""

    @staticmethod
    def compute_greeks(option, market: MarketData, d1, d2, is_call: bool) -> Greeks:
        backend = market.backend

        if option.expiry <= 0:
            if is_call:
                is_itm = backend.greater(market.spot, option.strike)
                delta = backend.where(is_itm, 1.0, 0.0)
            else:
                is_itm = backend.less(market.spot, option.strike)
                delta = backend.where(is_itm, -1.0, 0.0)
            zero = backend.zeros_like(delta) if hasattr(backend, "zeros_like") else 0.0
            return Greeks(delta=delta, gamma=zero, vega=zero, theta=zero, rho=zero)

        # Get adjusted parameters for Asian option
        vol_hat, drift_hat = compute_asian_adjusted_params(market)

        sqrt_T = backend.sqrt(backend.array(option.expiry))
        discount = compute_discount_factor(option.expiry, market)

        phi_d1 = _normal_pdf(d1, backend)
        N_d1 = backend.norm_cdf(d1)
        N_d2 = backend.norm_cdf(d2)

        # Adjusted spot price: S₀ * exp(drift_hat * T)
        adjusted_spot = backend.mul(market.spot, backend.exp(backend.mul(drift_hat, option.expiry)))

        if is_call:
            # Delta uses adjusted vol but original spot
            delta = backend.mul(
                discount, backend.mul(backend.exp(backend.mul(drift_hat, option.expiry)), N_d1)
            )

            # Rho calculation
            rho = backend.divide(
                backend.mul(backend.mul(option.strike, option.expiry), backend.mul(discount, N_d2)),
                100.0,
            )

            # Theta calculation
            theta_term1 = backend.divide(
                backend.mul(backend.mul(adjusted_spot, phi_d1), vol_hat), backend.mul(2.0, sqrt_T)
            )
            theta_term2 = backend.mul(
                backend.mul(market.rate, backend.mul(option.strike, discount)), N_d2
            )
            theta = backend.divide(
                backend.mul(-1.0, backend.mul(discount, backend.add(theta_term1, theta_term2))),
                365.0,
            )
        else:
            N_minus_d1 = backend.norm_cdf(backend.mul(-1.0, d1))
            N_minus_d2 = backend.norm_cdf(backend.mul(-1.0, d2))

            delta = backend.mul(
                -1.0,
                backend.mul(
                    discount,
                    backend.mul(backend.exp(backend.mul(drift_hat, option.expiry)), N_minus_d1),
                ),
            )

            rho = backend.divide(
                backend.mul(
                    -1.0,
                    backend.mul(
                        backend.mul(option.strike, option.expiry), backend.mul(discount, N_minus_d2)
                    ),
                ),
                100.0,
            )

            theta_term1 = backend.divide(
                backend.mul(backend.mul(adjusted_spot, phi_d1), vol_hat), backend.mul(2.0, sqrt_T)
            )
            theta_term2 = backend.mul(
                backend.mul(market.rate, backend.mul(option.strike, discount)), N_minus_d2
            )
            theta = backend.divide(
                backend.mul(discount, backend.add(backend.mul(-1.0, theta_term1), theta_term2)),
                365.0,
            )

        gamma = backend.divide(
            backend.mul(
                backend.mul(discount, phi_d1), backend.exp(backend.mul(drift_hat, option.expiry))
            ),
            backend.mul(backend.mul(market.spot, vol_hat), sqrt_T),
        )

        vega = backend.divide(
            backend.mul(backend.mul(backend.mul(discount, adjusted_spot), phi_d1), sqrt_T), 100.0
        )

        return Greeks(delta=delta, gamma=gamma, vega=vega, theta=theta, rho=rho)


class GeometricAsianCallFormula(PricingFormula):
    """Geometric Asian call pricing formula."""

    def price(self, option: GeometricAsianCall, market: MarketData):
        backend = market.backend
        vol_hat, drift_hat = compute_asian_adjusted_params(market)
        adjusted_market = market.with_vol(vol_hat)
        d1, d2 = compute_d1_d2(option.strike, option.expiry, adjusted_market, drift_hat)
        discount = compute_discount_factor(option.expiry, market)
        N_d1 = backend.norm_cdf(d1)
        N_d2 = backend.norm_cdf(d2)
        adjusted_spot = backend.mul(market.spot, backend.exp(backend.mul(drift_hat, option.expiry)))
        return backend.mul(
            discount,
            backend.add(backend.mul(adjusted_spot, N_d1), backend.mul(-option.strike, N_d2)),
        )

    def supports(self, option) -> bool:
        return isinstance(option, GeometricAsianCall)

    def greeks(self, option: GeometricAsianCall, market: MarketData, d1, d2) -> Greeks:
        return AsianGreeks.compute_greeks(option, market, d1, d2, is_call=True)

    def price_with_greeks(self, option: GeometricAsianCall, market: MarketData):
        vol_hat, drift_hat = compute_asian_adjusted_params(market)
        adjusted_market = market.with_vol(vol_hat)
        d1, d2 = compute_d1_d2(option.strike, option.expiry, adjusted_market, drift_hat)
        p = self.price(option, market)
        g = self.greeks(option, market, d1, d2)
        return p, g


class GeometricAsianPutFormula(PricingFormula):
    """Geometric Asian put pricing formula."""

    def price(self, option: GeometricAsianPut, market: MarketData):
        backend = market.backend
        vol_hat, drift_hat = compute_asian_adjusted_params(market)
        adjusted_market = market.with_vol(vol_hat)
        d1, d2 = compute_d1_d2(option.strike, option.expiry, adjusted_market, drift_hat)
        discount = compute_discount_factor(option.expiry, market)
        N_minus_d1 = backend.norm_cdf(backend.mul(-1.0, d1))
        N_minus_d2 = backend.norm_cdf(backend.mul(-1.0, d2))
        adjusted_spot = backend.mul(market.spot, backend.exp(backend.mul(drift_hat, option.expiry)))
        return backend.mul(
            discount,
            backend.add(
                backend.mul(option.strike, N_minus_d2), backend.mul(-adjusted_spot, N_minus_d1)
            ),
        )

    def supports(self, option) -> bool:
        return isinstance(option, GeometricAsianPut)

    def greeks(self, option: GeometricAsianPut, market: MarketData, d1, d2) -> Greeks:
        return AsianGreeks.compute_greeks(option, market, d1, d2, is_call=False)

    def price_with_greeks(self, option: GeometricAsianPut, market: MarketData):
        vol_hat, drift_hat = compute_asian_adjusted_params(market)
        adjusted_market = market.with_vol(vol_hat)
        d1, d2 = compute_d1_d2(option.strike, option.expiry, adjusted_market, drift_hat)
        p = self.price(option, market)
        g = self.greeks(option, market, d1, d2)
        return p, g



"""Core types and interfaces for Greeks calculation."""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Sequence
from enum import Enum, auto

from ...greeks import Greeks
from ...market import MarketData
from ...options.base import BaseLeg, OptionContract
from ...options.strategies import CompositeOptionContract

# Type alias for user convenience
Priceable = (
    OptionContract
    | BaseLeg
    | CompositeOptionContract
    | Sequence[OptionContract | BaseLeg | CompositeOptionContract]
)


class GreeksMethod(Enum):
    """Available methods for Greeks calculation in order of preference."""

    ANALYTICAL = auto()  # Fastest, limited coverage
    AUTODIFF = auto()  # Fast, broad coverage, requires JAX/torch
    FINITE_DIFF = auto()  # Slowest, universal fallback
    PRICER = auto()  # Use pricer's built-in Greeks (if available)

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented


class GreeksCalculator(ABC):
    """Abstract base for Greeks calculators."""

    def __init__(
        self,
        market: MarketData | None = None,
        *,
        spot: float | None = None,
        rate: float | None = None,
        vol: float | None = None,
    ):
        """Initialize calculator with optional market data or scalar overrides."""
        self._base_market = market
        self._base_spot = spot
        self._base_rate = rate
        self._base_vol = vol
        self.market = None

    def _create_market(self, backend):
        """Create market data with the given backend."""
        if self._base_market is not None:
            self.market = self._base_market
        else:
            self.market = MarketData.create(
                spot=self._base_spot if self._base_spot is not None else 100.0,
                rate=self._base_rate if self._base_rate is not None else 0.05,
                vol=self._base_vol if self._base_vol is not None else 0.2,
                backend=backend,
            )

    @abstractmethod
    def supports(self, option: Priceable) -> bool:
        """Check if this calculator supports the given option type."""
        pass

    @abstractmethod
    def calculate(
        self,
        option: Priceable,
        *,
        market: MarketData | None = None,
        spot: float | None = None,
        rate: float | None = None,
        vol: float | None = None,
    ) -> Greeks:
        """Calculate Greeks for option or portfolio."""
        pass

    @abstractmethod
    def calculate_greeks(
        self,
        option: OptionContract | BaseLeg | CompositeOptionContract,
    ) -> Greeks:
        """Calculate Greeks for the given option."""
        pass

    def _resolve_market(
        self,
        provided: MarketData | None,
        spot: float | None,
        rate: float | None,
        vol: float | None,
    ) -> MarketData:
        """Resolve final MarketData from optional overrides."""
        m = provided or self.market
        if spot is not None:
            m = m.with_spot(spot)
        if rate is not None:
            m = m.with_rate(rate)
        if vol is not None:
            m = m.with_vol(vol)
        return m

    def _zero_greeks(
        self,
        vol: object | None = None,
        spot: object | None = None,
        rate: object | None = None,
    ) -> Greeks:
        """Return zero Greeks with appropriate shape."""
        params = [p for p in (vol, spot, rate) if p is not None]
        arrays = [p for p in params if hasattr(p, "shape")]

        if arrays:
            ref = arrays[0]
            zero = ref * 0
            return Greeks(delta=zero, gamma=zero, vega=zero, theta=zero, rho=zero)

        return Greeks()



"""Fallback backend for when no array libraries are installed."""

from ..interface import AbstractBackend


class FallbackBackend(AbstractBackend):
    """Fallback backend that raises helpful errors."""

    name = "fallback"

    def __init__(self):
        self._error_msg = (
            "No array backend available. Install one of:\n"
            "  pip install numpy\n"
            "  pip install jax jaxlib\n"
            "  pip install torch"
        )

    def _not_available(self, *args, **kwargs):
        raise RuntimeError(self._error_msg)

    def __getattribute__(self, name):
        if name in (
            "name",
            "_error_msg",
            "_not_available",
            "__class__",
            "__init__",
        ):
            return object.__getattribute__(self, name)
        return object.__getattribute__(self, "_not_available")

    # Add explicit methods to satisfy the protocol
    def gather(self, a, indices, axis=0):
        return self._not_available()

    def norm(self, a, ord=None, axis=None):
        return self._not_available()

    def solve(self, a, b):
        return self._not_available()

    def inv(self, a):
        return self._not_available()

    def det(self, a):
        return self._not_available()



"""Base sampler for distributions with abstract backend support."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any

from ...backend import ArrayBackend, active_backend
from ...types import Shape


class BaseSampler(ABC):
    """Base sampler for distributions with abstract backend support."""

    def __init__(self, mean: float = 0.0, std: float = 1.0):
        """Initialize base sampler.

        Args:
            mean: Mean of the distribution
            std: Standard deviation of the distribution

        Raises:
            ValueError: If std <= 0
        """
        if std <= 0:
            raise ValueError(f"Standard deviation must be positive, got {std}")
        self._mean = float(mean)
        self._std = float(std)
        self._mgf_cache: dict[float, float] = {}
        self._skew: float | None = None
        self._ex_kurtosis: float | None = None

    @abstractmethod
    def _raw_sample(self, shape: Shape, backend: ArrayBackend, key: Any = None) -> Any:
        """Draw from standardized distribution using backend.

        Args:
            shape: Shape of samples to draw
            backend: Backend to use for sampling
            key: Random key (for JAX backend)

        Returns:
            Array of samples from standardized distribution
        """
        pass

    @abstractmethod
    def sample(
        self,
        shape: Shape,
        backend: ArrayBackend | None = None,
        key: Any = None,
        standardize: bool = False,
    ) -> Any:
        """Sample from distribution using backend.

        Args:
            shape: Shape of samples to draw
            backend: Backend to use (defaults to active backend)
            key: Random key (for JAX backend)
            standardize: If True, return standardized samples

        Returns:
            Array of samples
        """
        pass

    @abstractmethod
    def pdf(
        self,
        x: Any,
        backend: ArrayBackend | None = None,
        standardize: bool = False,
    ) -> Any:
        """Compute PDF at x using backend.

        Args:
            x: Points at which to evaluate PDF
            backend: Backend to use (defaults to active backend)
            standardize: If True, compute standardized PDF

        Returns:
            PDF values at x
        """
        pass

    @abstractmethod
    def cdf(
        self,
        x: Any,
        backend: ArrayBackend | None = None,
        standardize: bool = False,
    ) -> Any:
        """Compute CDF at x using backend.

        Args:
            x: Points at which to evaluate CDF
            backend: Backend to use (defaults to active backend)
            standardize: If True, compute standardized CDF

        Returns:
            CDF values at x
        """
        pass

    @abstractmethod
    def ppf(
        self,
        q: Any,
        backend: ArrayBackend | None = None,
        standardize: bool = False,
    ) -> Any:
        """Compute inverse CDF (quantile function) at q.

        Args:
            q: Quantiles (must be in [0, 1])
            backend: Backend to use (defaults to active backend)
            standardize: If True, compute standardized quantiles

        Returns:
            Quantile values
        """
        pass

    def _compute_moments(self, backend: ArrayBackend | None = None) -> tuple[float, float]:
        """Compute skew and excess kurtosis using backend.

        Args:
            backend: Backend to use (defaults to active backend)

        Returns:
            Tuple of (skew, excess_kurtosis)
        """
        if backend is None:
            backend = active_backend()

        n = 100_000
        if backend.name == "jax":
            key = backend.random_key(42)
            z = self._raw_sample((n,), backend=backend, key=key)
        else:
            z = self._raw_sample((n,), backend=backend)

        z_std = (z - self._mean) / self._std
        m3 = backend.mean(backend.pow(z_std, 3))
        m4 = backend.mean(backend.pow(z_std, 4))

        skew = float(backend.to_numpy(m3))
        ex_kurt = float(backend.to_numpy(m4)) - 3.0

        return skew, ex_kurt

    def mgf(self, t: float, backend: ArrayBackend | None = None) -> float:
        """Estimate MGF using backend with numerical stability.

        Args:
            t: Point at which to evaluate MGF
            backend: Backend to use (defaults to active backend)

        Returns:
            MGF value at t

        Raises:
            ValueError: If MGF is unstable or invalid
        """
        if backend is None:
            backend = active_backend()

        key = round(t, 8)
        if key not in self._mgf_cache:
            n = 100_000

            if backend.name == "jax":
                rng_key = backend.random_key(42)
                samples = self.sample((n,), backend=backend, key=rng_key)
            else:
                samples = self.sample((n,), backend=backend)

            tx = backend.mul(backend.array(t), samples)
            max_tx = backend.max(tx)

            exp_normalized = backend.exp(backend.sub(tx, max_tx))
            mgf_val = backend.mul(backend.exp(max_tx), backend.mean(exp_normalized))
            mgf_float = float(backend.to_numpy(mgf_val))

            if not backend.to_numpy(backend.isfinite(mgf_val)) or mgf_float <= 0:
                raise ValueError(f"MGF unstable for t={t}: got {mgf_float}")

            self._mgf_cache[key] = mgf_float

        return self._mgf_cache[key]

    @property
    def mean(self) -> float:
        """Mean of the distribution."""
        return self._mean

    @property
    def std(self) -> float:
        """Standard deviation of the distribution."""
        return self._std

    @property
    def variance(self) -> float:
        """Variance of the distribution."""
        return self._std**2

    @property
    def skew(self) -> float:
        """Skewness of the distribution."""
        if self._skew is None:
            self._skew, self._ex_kurtosis = self._compute_moments()
        return self._skew

    @property
    def ex_kurtosis(self) -> float:
        """Excess kurtosis of the distribution."""
        if self._ex_kurtosis is None:
            self._skew, self._ex_kurtosis = self._compute_moments()
        return self._ex_kurtosis

    @abstractmethod
    def __repr__(self) -> str:
        """String representation of the sampler."""
        pass



"""Asian option contracts."""

from dataclasses import dataclass

from ..utils import infer_backend
from .base import PathDependentOption, PayoffType, Side


@dataclass(frozen=True)
class AsianCall(PathDependentOption):
    """Asian call option contract."""

    @property
    def payoff_type(self) -> PayoffType:
        return PayoffType.ASIAN

    @property
    def side(self) -> Side:
        return Side.CALL

    @infer_backend
    def _path_payoff(self, price_paths, backend=None):
        """Compute Asian call payoff: max(Avg(S) - K, 0)."""
        price_paths = backend.array(price_paths)
        average_price = backend.mean(price_paths, axis=1)
        strike = backend.array(self.strike)
        zero = backend.array(0.0)
        return backend.maximum(backend.add(average_price, backend.mul(-1, strike)), zero)


@dataclass(frozen=True)
class AsianPut(PathDependentOption):
    """Asian put option contract."""

    @property
    def payoff_type(self) -> PayoffType:
        return PayoffType.ASIAN

    @property
    def side(self) -> Side:
        return Side.PUT

    @infer_backend
    def _path_payoff(self, price_paths, backend=None):
        """Compute Asian put payoff: max(K - Avg(S), 0)."""
        price_paths = backend.array(price_paths)
        average_price = backend.mean(price_paths, axis=1)
        strike = backend.array(self.strike)
        zero = backend.array(0.0)
        return backend.maximum(backend.add(strike, backend.mul(-1, average_price)), zero)


@dataclass(frozen=True)
class GeometricAsianCall(PathDependentOption):
    """Geometric Asian call option contract."""

    @property
    def payoff_type(self) -> PayoffType:
        return PayoffType.GEOMETRIC_ASIAN

    @property
    def side(self) -> Side:
        return Side.CALL

    @infer_backend
    def _path_payoff(self, price_paths, backend=None):
        """Compute geometric Asian call payoff: max(GeoAvg(S) - K, 0)."""
        price_paths = backend.array(price_paths)
        geometric_average = backend.exp(backend.mean(backend.log(price_paths), axis=1))
        strike = backend.array(self.strike)
        zero = backend.array(0.0)
        return backend.maximum(backend.add(geometric_average, backend.mul(-1, strike)), zero)


@dataclass(frozen=True)
class GeometricAsianPut(PathDependentOption):
    """Geometric Asian put option contract."""

    @property
    def payoff_type(self) -> PayoffType:
        return PayoffType.GEOMETRIC_ASIAN

    @property
    def side(self) -> Side:
        return Side.PUT

    @infer_backend
    def _path_payoff(self, price_paths, backend=None):
        """Compute geometric Asian put payoff: max(K - GeoAvg(S), 0)."""
        price_paths = backend.array(price_paths)
        geometric_average = backend.exp(backend.mean(backend.log(price_paths), axis=1))
        strike = backend.array(self.strike)
        zero = backend.array(0.0)
        return backend.maximum(backend.add(strike, backend.mul(-1, geometric_average)), zero)



"""Digital option contracts."""

from dataclasses import dataclass

from ..utils import infer_backend
from .base import PathIndependentOption, PayoffType, Side


@dataclass(frozen=True)
class DigitalCall(PathIndependentOption):
    """Digital call option contract."""

    payout: float = 1.0

    @property
    def payoff_type(self) -> PayoffType:
        return PayoffType.DIGITAL

    @property
    def side(self) -> Side:
        return Side.CALL

    @infer_backend
    def payoff(self, spot_prices, backend=None):
        spot_prices = backend.array(spot_prices)
        strike = backend.array(self.strike)
        cond = backend.greater(spot_prices, strike)
        payout = backend.array(self.payout)
        zero = backend.array(0.0)
        return backend.where(cond, payout, zero)

    def intrinsic_value(self, spot: float) -> float:
        return self.payout if spot > self.strike else 0.0


@dataclass(frozen=True)
class DigitalPut(PathIndependentOption):
    """Digital put option contract."""

    payout: float = 1.0

    @property
    def payoff_type(self) -> PayoffType:
        return PayoffType.DIGITAL

    @property
    def side(self) -> Side:
        return Side.PUT

    @infer_backend
    def payoff(self, spot_prices, backend=None):
        spot_prices = backend.array(spot_prices)
        strike = backend.array(self.strike)
        cond = backend.less(spot_prices, strike)
        payout = backend.array(self.payout)
        zero = backend.array(0.0)
        return backend.where(cond, payout, zero)

    def intrinsic_value(self, spot: float) -> float:
        return self.payout if spot < self.strike else 0.0



"""Monte Carlo pricing engine."""

from __future__ import annotations

from math import ceil

from laakhay.quantlab.backend import get_backend

from ..market import MarketData
from ..simulations.models import GeometricBrownianMotionSimulation
from ..simulations.samplers import NormalSampler


class MonteCarloPricer:
    """Monte Carlo engine for option pricing."""

    def __init__(
        self,
        market: MarketData | None = None,
        sampler=None,
        n_paths: int = 20_000,
        steps_per_year: int = 252,
        seed: int | None = None,
        antithetic: bool = True,
        cache: bool = True,
        moment_match: bool = True,
        stratify: bool = False,
        backend=None,
    ):
        self.backend = backend or get_backend()
        self.default_market = market or MarketData(backend=self.backend)

        self.n_paths = n_paths
        self.steps_per_year = steps_per_year
        self.seed = seed
        self.antithetic = antithetic
        self.cache = cache
        self.moment_match = moment_match
        self.stratify = stratify

        self.sampler = sampler or NormalSampler()
        self.simulator = GeometricBrownianMotionSimulation(
            sampler=self.sampler,
            antithetic=self.antithetic,
            moment_match=self.moment_match,
            stratify=self.stratify,
        )

        self._path_cache = {}

    def get_paths(self, expiry: float, market: MarketData | None = None):
        """Generate price paths."""
        market = market or self.default_market
        n_steps = max(1, int(ceil(self.steps_per_year * expiry)))

        # Simple caching based on expiry and n_steps
        def _make_hashable(val):
            try:
                hash(val)
                return val
            except TypeError:
                # If not hashable, it's likely an array/tensor
                # We convert to a tuple of its flattened values
                return tuple(self.backend.to_numpy(val).flatten().tolist())

        cache_key = (
            expiry,
            n_steps,
            _make_hashable(market.spot),
            _make_hashable(market.rate),
            _make_hashable(market.vol),
        )
        if self.cache and cache_key in self._path_cache:
            return self._path_cache[cache_key]

        # Set seed if provided
        if self.seed is not None:
            self.backend.random_key(self.seed)

        paths = self.simulator.generate_paths(
            n_paths=self.n_paths,
            n_steps=n_steps,
            expiry=expiry,
            spot=market.spot,
            rate=market.rate,
            vol=market.vol,
            backend=self.backend,
        )

        if self.cache:
            self._path_cache[cache_key] = paths

        return paths

    def price(self, option, market: MarketData | None = None) -> object:
        """Price option using Monte Carlo."""
        market = market or self.default_market

        paths = self.get_paths(option.expiry, market)

        # Payoff calculation
        payoffs = option(paths, backend=self.backend)
        discount = self.backend.exp(
            self.backend.mul(self.backend.mul(-1, market.rate), option.expiry)
        )

        return self.backend.mul(discount, self.backend.mean(payoffs, axis=0))

    def supports(self, option) -> bool:
        """Check if option is supported."""
        from ..options.base import OptionContract

        return isinstance(option, OptionContract)



from __future__ import annotations

from decimal import Decimal
from typing import Any


class PositionSizer:
    """Calculates position size based on risk parameters."""

    def __init__(self, capital_provider: callable):
        """
        Args:
            capital_provider: Function that returns current capital.
        """
        self._get_capital = capital_provider

    def calculate_size(self, signal: Any, price: Any) -> Decimal:
        """Calculate quantity based on risk parameters."""
        px = Decimal(str(price))
        if px <= 0:
            return Decimal("0")

        # 1. Check Signal override
        if signal.size:
            return self._parse_size(signal.size, px)

        # 2. Default Strategy Sizing (Assuming mostly fixed for now or hardcoded default)
        # TODO: Get defaults from Strategy proper
        return self._parse_size("10%", px)  # Default 10% equity

    def _parse_size(self, size: str | Decimal | float, price: Decimal) -> Decimal:
        """Parse size string ("1%", "100") or number."""
        capital = self._get_capital()
        if capital <= 0 or price <= 0:
            return Decimal("0")

        # Always cap by what can be bought with available cash.
        max_affordable_qty = capital / price

        if isinstance(size, int | float | Decimal):
            # If plain number, assume Quantity? Or Fixed Cash?
            # Ambiguous. Let's assume Fixed Cash for float/int/decimal unless small?
            # Creating a convention: < 1.0 is %, > 1.0 is Cash.
            # This is risky. Better to stick to string for %
            val = Decimal(str(size))
            if val <= 0:
                return Decimal("0")
            qty = (capital * val) / price if val < 1 else min(val, capital) / price
            return min(qty, max_affordable_qty)

        if isinstance(size, str):
            if size.endswith("%"):
                pct = Decimal(size[:-1]) / 100
                if pct <= 0:
                    return Decimal("0")
                qty = (capital * pct) / price
                return min(qty, max_affordable_qty)
            else:
                cash = Decimal(size)
                if cash <= 0:
                    return Decimal("0")
                return min(cash, capital) / price

        return Decimal("0")



"""Unified Greeks calculation engine with automatic method selection."""

from __future__ import annotations

from laakhay.quantlab.backend import Backend, get_backend
from laakhay.quantlab.pricing.greeks import Greeks
from laakhay.quantlab.pricing.market import MarketData

from ...options.base import BaseLeg, OptionContract
from ...options.strategies import CompositeOptionContract
from ..black_scholes import BlackScholesPricer
from .base import GreeksCalculator, GreeksMethod, Priceable
from .utils import resolve_market_overrides


class GreeksEngine(GreeksCalculator):
    """Smart Greeks calculator that selects the best available method."""

    def __init__(
        self,
        backend: str | Backend | None = None,
        market: MarketData | None = None,
        preferred_methods: list[GreeksMethod] | None = None,
        *,
        spot: float | None = None,
        rate: float | None = None,
        vol: float | None = None,
    ) -> None:
        """Initialize with optional method preferences."""
        super().__init__(market=market, spot=spot, rate=rate, vol=vol)
        self.backend = get_backend(backend)
        self._create_market(self.backend)

        # Default method order
        self._methods = preferred_methods or [
            GreeksMethod.PRICER,
            GreeksMethod.AUTODIFF,
            GreeksMethod.FINITE_DIFF,
        ]

        # Initialize calculators lazily
        self._analytical = None
        self._autodiff = None
        self._finite_diff = None
        self._pricer = BlackScholesPricer(backend=self.backend)

        # Cache method selection
        self._method_cache: dict[type[OptionContract], GreeksMethod] = {}

    def supports(self, option: Priceable) -> bool:
        """Check if any method supports this option type."""
        return self._get_best_method(option) is not None

    def _get_best_method(self, option: Priceable) -> GreeksMethod | None:
        """Select best available method for option type."""
        if isinstance(option, (list, tuple, set)):
            methods = [self._get_best_method(opt) for opt in option]
            if None in methods:
                return None
            return max(methods)  # type: ignore

        if isinstance(option, OptionContract):
            cached = self._method_cache.get(type(option))
            if cached is not None:
                return cached

        for method in self._methods:
            if method == GreeksMethod.PRICER:
                if hasattr(self._pricer, "price_with_greeks") and self._pricer.supports(option):
                    if isinstance(option, OptionContract):
                        self._method_cache[type(option)] = method
                    return method
            elif method == GreeksMethod.AUTODIFF:
                ad = self._get_autodiff()
                if ad and ad.supports(option):
                    if isinstance(option, OptionContract):
                        self._method_cache[type(option)] = method
                    return method
            elif method == GreeksMethod.FINITE_DIFF:
                fd = self._get_finite_diff()
                if fd and fd.supports(option):
                    if isinstance(option, OptionContract):
                        self._method_cache[type(option)] = method
                    return method

        return None

    def _get_autodiff(self):
        """Get or create autodiff calculator."""
        if self._autodiff is None:
            try:
                from .autodiff import AutoDiffGreeksCalculator

                self._autodiff = AutoDiffGreeksCalculator(
                    backend=self.backend,
                    market=self.market,
                )
            except ImportError:
                return None
        return self._autodiff

    def _get_finite_diff(self):
        """Get or create finite difference calculator."""
        if self._finite_diff is None:
            try:
                from .finite_diff import FiniteDiffGreeksCalculator

                self._finite_diff = FiniteDiffGreeksCalculator(
                    backend=self.backend,
                    market=self.market,
                )
            except ImportError:
                return None
        return self._finite_diff

    def calculate_greeks(
        self,
        option: OptionContract | BaseLeg | CompositeOptionContract,
    ) -> Greeks:
        """Calculate Greeks for the given option."""
        return self.calculate(option)

    def calculate(
        self,
        option: Priceable,
        *,
        market: MarketData | None = None,
        spot: float | None = None,
        rate: float | None = None,
        vol: float | None = None,
        method: str | None = None,
    ) -> Greeks:
        """Calculate Greeks with automatic method selection."""
        final_market = resolve_market_overrides(self.market, market, spot, rate, vol)

        if isinstance(option, (list, tuple, set)):
            total_greeks = None
            for single_option in option:
                single_greeks = self.calculate(single_option, market=final_market, method=method)
                if total_greeks is None:
                    total_greeks = single_greeks
                else:
                    total_greeks = total_greeks + single_greeks
            return total_greeks or Greeks()

        if method is None or method.lower() in {"auto", "unified"}:
            best_method = self._get_best_method(option)
            if best_method is None:
                raise ValueError(f"No Greeks method supports {type(option)}")

            if best_method == GreeksMethod.PRICER:
                _, greeks = self._pricer.price_with_greeks(option, final_market)
                return greeks
            elif best_method == GreeksMethod.AUTODIFF:
                return self._get_autodiff().calculate(option, market=final_market)
            elif best_method == GreeksMethod.FINITE_DIFF:
                return self._get_finite_diff().calculate(option, market=final_market)

        # Explicit method
        method_norm = method.lower()
        if method_norm in {"analytical", "analytic", "pricer"}:
            _, greeks = self._pricer.price_with_greeks(option, final_market)
            return greeks
        elif method_norm in {"autodiff", "ad"}:
            calc = self._get_autodiff()
            if calc is None:
                raise RuntimeError("Autodiff not available")
            return calc.calculate(option, market=final_market)
        elif method_norm in {"finite_diff", "fd"}:
            calc = self._get_finite_diff()
            if calc is None:
                raise RuntimeError("Finite diff not available")
            return calc.calculate(option, market=final_market)
        else:
            raise ValueError(f"Unknown method '{method}'")



"""NumPy backend."""

from __future__ import annotations

from ...types import Array, ArrayLike, Axis, Shape
from ..interface import AbstractBackend

try:
    import numpy as np

    HAS_NUMPY = True
except ImportError:
    HAS_NUMPY = False
    np = None


class NumpyBackend(AbstractBackend):
    """NumPy backend."""

    name = "numpy"

    def __init__(self) -> None:
        if not HAS_NUMPY:
            raise ImportError("NumPy not installed. Run: pip install numpy")

    def array(self, data: ArrayLike, dtype: any = None, device: any = None) -> Array:
        return np.array(data, dtype=dtype)

    def zeros(self, shape: Shape, dtype: any = None, device: any = None) -> Array:
        return np.zeros(shape, dtype=dtype)

    def ones(self, shape: Shape, dtype: any = None, device: any = None) -> Array:
        return np.ones(shape, dtype=dtype)

    def arange(self, *args, dtype: any = None, device: any = None) -> Array:
        return np.arange(*args, dtype=dtype)

    def linspace(self, start: float, stop: float, num: int = 50, device: any = None) -> Array:
        return np.linspace(start, stop, num)

    def eye(self, n: int, m: int | None = None, device: any = None) -> Array:
        return np.eye(n, m)

    def reshape(self, a: Array, shape: Shape) -> Array:
        return np.reshape(a, shape)

    def transpose(self, a: Array, axes: list[int] | None = None) -> Array:
        return np.transpose(a, axes)

    def concat(self, arrays: list[Array], axis: int = 0) -> Array:
        return np.concatenate(arrays, axis)

    def stack(self, arrays: list[Array], axis: int = 0) -> Array:
        return np.stack(arrays, axis)

    def split(self, a: Array, indices: int | list[int], axis: int = 0) -> list[Array]:
        return np.split(a, indices, axis)

    def sum(self, a: Array, axis: Axis = None, keepdims: bool = False) -> Array:
        return np.sum(a, axis=axis, keepdims=keepdims)

    def mean(self, a: Array, axis: Axis = None, keepdims: bool = False) -> Array:
        return np.mean(a, axis=axis, keepdims=keepdims)

    def std(self, a: Array, axis: Axis = None, keepdims: bool = False) -> Array:
        return np.std(a, axis=axis, keepdims=keepdims)

    def min(self, a: Array, axis: Axis = None, keepdims: bool = False) -> Array:
        return np.min(a, axis=axis, keepdims=keepdims)

    def max(self, a: Array, axis: Axis = None, keepdims: bool = False) -> Array:
        return np.max(a, axis=axis, keepdims=keepdims)

    def argmin(self, a: Array, axis: Axis = None) -> Array:
        return np.argmin(a, axis=axis)

    def argmax(self, a: Array, axis: Axis = None) -> Array:
        return np.argmax(a, axis=axis)

    def pow(self, a: Array, b: Array) -> Array:
        return np.power(a, b)

    def matmul(self, a: Array, b: Array) -> Array:
        return np.matmul(a, b)

    def abs(self, a: Array) -> Array:
        return np.abs(a)

    def sign(self, a: Array) -> Array:
        return np.sign(a)

    def exp(self, a: Array) -> Array:
        return np.exp(a)

    def log(self, a: Array) -> Array:
        return np.log(a)

    def sqrt(self, a: Array) -> Array:
        return np.sqrt(a)

    def sin(self, a: Array) -> Array:
        return np.sin(a)

    def cos(self, a: Array) -> Array:
        return np.cos(a)

    def tanh(self, a: Array) -> Array:
        return np.tanh(a)

    def where(self, cond: Array, x: Array, y: Array) -> Array:
        return np.where(cond, x, y)

    def clip(self, a: Array, min: float | None = None, max: float | None = None) -> Array:
        return np.clip(a, min, max)

    def maximum(self, a: Array, b: Array) -> Array:
        return np.maximum(a, b)

    def minimum(self, a: Array, b: Array) -> Array:
        return np.minimum(a, b)

    def shape(self, a: Array) -> Shape:
        return a.shape

    def dtype(self, a: Array) -> any:
        return a.dtype

    def ndim(self, a: Array) -> int:
        return a.ndim

    def size(self, a: Array) -> int:
        return a.size

    def is_array(self, obj: any) -> bool:
        return isinstance(obj, np.ndarray)

    def item(self, a: Array) -> float | int | bool:
        return a.item()

    def to_numpy(self, a: Array) -> Array:
        return a

    def from_numpy(self, a: Array) -> Array:
        return a

    def cast(self, a: Array, dtype: any) -> Array:
        return a.astype(dtype)

    def eq(self, a: Array, b: Array) -> Array:
        return np.equal(a, b)

    def ne(self, a: Array, b: Array) -> Array:
        return np.not_equal(a, b)

    def lt(self, a: Array, b: Array) -> Array:
        return np.less(a, b)

    def le(self, a: Array, b: Array) -> Array:
        return np.less_equal(a, b)

    def gt(self, a: Array, b: Array) -> Array:
        return np.greater(a, b)

    def ge(self, a: Array, b: Array) -> Array:
        return np.greater_equal(a, b)

    def isnan(self, a: Array) -> Array:
        return np.isnan(a)

    def isinf(self, a: Array) -> Array:
        return np.isinf(a)

    def isfinite(self, a: Array) -> Array:
        return np.isfinite(a)

    def add(self, a: Array, b: Array) -> Array:
        return np.add(a, b)

    def sub(self, a: Array, b: Array) -> Array:
        return np.subtract(a, b)

    def mul(self, a: Array, b: Array) -> Array:
        return np.multiply(a, b)

    def div(self, a: Array, b: Array) -> Array:
        return np.divide(a, b)

    def dot(self, a: Array, b: Array) -> Array:
        return np.dot(a, b)

    def tan(self, a: Array) -> Array:
        return np.tan(a)

    def sinh(self, a: Array) -> Array:
        return np.sinh(a)

    def cosh(self, a: Array) -> Array:
        return np.cosh(a)

    def log10(self, a: Array) -> Array:
        return np.log10(a)

    def var(self, a: Array, axis: Axis = None, keepdims: bool = False) -> Array:
        return np.var(a, axis=axis, keepdims=keepdims)

    def prod(self, a: Array, axis: Axis = None, keepdims: bool = False) -> Array:
        return np.prod(a, axis=axis, keepdims=keepdims)

    def squeeze(self, a: Array, axis: Axis = None) -> Array:
        return np.squeeze(a, axis=axis)

    def expand_dims(self, a: Array, axis: int) -> Array:
        return np.expand_dims(a, axis=axis)

    def cumsum(self, a: Array, axis: int | None = None) -> Array:
        return np.cumsum(a, axis=axis)

    def any(self, a: Array, axis: Axis = None, keepdims: bool = False) -> Array:
        return np.any(a, axis=axis, keepdims=keepdims)

    def all(self, a: Array, axis: Axis = None, keepdims: bool = False) -> Array:
        return np.all(a, axis=axis, keepdims=keepdims)

    def logical_not(self, a: Array) -> Array:
        return np.logical_not(a)

    def logical_and(self, a: Array, b: Array) -> Array:
        return np.logical_and(a, b)

    def logical_or(self, a: Array, b: Array) -> Array:
        return np.logical_or(a, b)

    def take(self, a: Array, indices: Array, axis: int | None = None) -> Array:
        return np.take(a, indices, axis=axis)

    def is_gpu_available(self) -> bool:
        return False

    def device(self, array: Array) -> any:
        from ..device import Device

        return Device.cpu()

    def to_device(self, array: Array, device: any) -> Array:
        return array

    def gather(self, a: Array, indices: Array, axis: int = 0) -> Array:
        return np.take(a, indices, axis=axis)

    def norm(self, a: Array, ord: any = None, axis: Axis = None) -> Array:
        return np.linalg.norm(a, ord=ord, axis=axis)

    def solve(self, a: Array, b: Array) -> Array:
        return np.linalg.solve(a, b)

    def inv(self, a: Array) -> Array:
        return np.linalg.inv(a)

    def det(self, a: Array) -> Array:
        return np.linalg.det(a)

    def erf(self, x: Array) -> Array:
        try:
            from scipy.special import erf

            return erf(x)
        except ImportError:
            import numpy as _np

            a1 = 0.254829592
            a2 = -0.284496736
            a3 = 1.421413741
            a4 = -1.453152027
            a5 = 1.061405429
            p = 0.3275911
            sign = _np.sign(x)
            x = _np.abs(x)
            t = 1.0 / (1.0 + p * x)
            y = 1.0 - (((((a5 * t + a4) * t) + a3) * t + a2) * t + a1) * t * _np.exp(-x * x)
            return sign * y

    def erfc(self, x: Array) -> Array:
        from scipy.special import erfc

        return erfc(x)

    def norm_cdf(self, x: Array) -> Array:
        try:
            from scipy.stats import norm

            return norm.cdf(x)
        except ImportError:
            return 0.5 * (1 + self.erf(x / np.sqrt(2)))

    def norm_pdf(self, x: Array) -> Array:
        from scipy.stats import norm

        return norm.pdf(x)

    def norm_ppf(self, q: Array) -> Array:
        try:
            from scipy.stats import norm

            return norm.ppf(q)
        except ImportError:
            raise NotImplementedError("norm_ppf requires scipy for numpy backend") from None

    def gamma(self, x: Array) -> Array:
        from scipy.special import gamma

        return gamma(x)

    def lgamma(self, x: Array) -> Array:
        from scipy.special import gammaln

        return gammaln(x)

    def random_key(self, seed: int) -> any:
        np.random.seed(seed)
        return seed

    def random_normal(
        self,
        key: any,
        shape: Shape,
        dtype: any = None,
        device: any = None,
    ) -> Array:
        result = np.random.normal(0.0, 1.0, shape)
        if dtype is not None:
            result = result.astype(dtype)
        return result

    def random_uniform(
        self,
        key: any,
        shape: Shape,
        dtype: any = None,
        device: any = None,
        low: float = 0.0,
        high: float = 1.0,
    ) -> Array:
        result = np.random.uniform(low, high, shape)
        if dtype is not None:
            result = result.astype(dtype)
        return result

    def random_split(self, key: any, num: int = 2) -> list[any]:
        return [key + i for i in range(num)]



"""Option contracts and strategy definitions."""

from .asian import AsianCall, AsianPut, GeometricAsianCall, GeometricAsianPut
from .barrier import (
    BarrierDirection,
    BarrierOption,
    DownAndInCall,
    DownAndInPut,
    DownAndOutCall,
    DownAndOutPut,
    KnockType,
    UpAndInCall,
    UpAndInPut,
    UpAndOutCall,
    UpAndOutPut,
)
from .base import (
    CashLeg,
    OptionContract,
    OptionLeg,
    PathDependentOption,
    PathIndependentOption,
    PayoffType,
    PositionType,
    Side,
)
from .contracts import EuropeanCall, EuropeanPut
from .digital import DigitalCall, DigitalPut
from .strategies import (
    CompositeOptionContract,
    butterfly_spread,
    calendar_spread,
    iron_condor,
    straddle,
    strangle,
    vertical_spread,
)

__all__ = [
    "Side",
    "PayoffType",
    "PositionType",
    "OptionContract",
    "PathIndependentOption",
    "PathDependentOption",
    "OptionLeg",
    "CashLeg",
    "EuropeanCall",
    "EuropeanPut",
    "DigitalCall",
    "DigitalPut",
    "AsianCall",
    "AsianPut",
    "GeometricAsianCall",
    "GeometricAsianPut",
    "BarrierOption",
    "BarrierDirection",
    "KnockType",
    "UpAndOutCall",
    "UpAndInCall",
    "DownAndOutCall",
    "DownAndInCall",
    "UpAndOutPut",
    "UpAndInPut",
    "DownAndOutPut",
    "DownAndOutPut",
    "DownAndInPut",
    "CompositeOptionContract",
    "vertical_spread",
    "calendar_spread",
    "butterfly_spread",
    "straddle",
    "strangle",
    "iron_condor",
]



"""Simulation tools for quantitative finance."""

from .gbm import GeometricBrownianMotion
from .samplers import BaseSampler, GaussianSampler

__all__ = [
    "BaseSampler",
    "GaussianSampler",
    "GeometricBrownianMotion",
]
